;;; analyze.el --- run analyzer as inferior of Emacs, parse error messages.

;; Copyright (C) 1985, 1986, 1987, 1993, 1994 Free Software Foundation, Inc.

;; Author: Roland McGrath <roland@prep.ai.mit.edu>
;; Maintainer: FSF
;; Keywords: tools, processes

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Code:

;;;###autoload
(defvar analysis-mode-hook nil
  "*List of hook functions run by `analysis-mode' (see `run-hooks').")


;;;###autoload
(defconst analysis-window-height nil
  "*Number of lines in a analysis window.  If nil, use Emacs default.")


(defvar analysis-error-list nil
  "List of error message descriptors for visiting erring functions.
Each error descriptor is a cons (or nil).  Its car is a marker pointing to
an error message.  If its cdr is a marker, it points to the text of the
line the message is about.  If its cdr is a cons, that cons's car is a cons
\(DIRECTORY . FILE\), specifying the file the message is about, and its cdr
is the number of the line the message is about.  Or its cdr may be nil if
that error is not interesting.

The value may be t instead of a list; this means that the buffer of
error messages should be reparsed the next time the list of errors is wanted.

Some other commands (like `diff') use this list to control the error
message tracking facilites; if you change its structure, you should make
sure you also change those packages.  Perhaps it is better not to change
it at all.")


(defvar analysis-old-error-list nil
  "Value of `analysis-error-list' after errors were parsed.")


(defvar analysis-parse-errors-function 'analysis-parse-errors
  "Function to call to parse error messages from a analysis.
   It takes args LIMIT-SEARCH and FIND-AT-LEAST.
   If LIMIT-SEARCH is non-nil, don't bother parsing past that location.
   If FIND-AT-LEAST is non-nil, don't bother parsing after finding that
   many new erros.
   It should read in the source files which have errors and set
   `analysis-error-list' to a list with an element for each error message
   found.  See that variable for more info.")


;;;###autoload
(defvar analysis-buffer-name-function nil
  "Function to compute the name of a analysis buffer.
   The function receives one argument, the name of the major mode of the
   analysis buffer.  It should return a string.
   nil means use default '*Analysis-Results*'.")


;;;###autoload
(defvar analysis-finish-function nil
  "*Function to call when a analysis process finishes.
   It is called with two arguments: the analysis buffer, and a string
   describing how the process finished.")


(defvar analysis-last-buffer nil
  "The most recent analysis buffer.
   A buffer becomes most recent when its analysis is started
   or when it is used with \\[analyze-next-error] or \\[analyze-goto-error].")


(defvar analysis-in-progress nil
  "List of analysis processes now running.")

(or (assq 'analysis-in-progress minor-mode-alist)
    (setq minor-mode-alist (cons '(analysis-in-progress " Analyzing")
				 minor-mode-alist)))


(defvar analysis-parsing-end nil
  "Position of end of buffer when last error messages were parsed.")


(defvar analysis-error-message "No more errors"
  "Message to print when no more matches are found.")


(defvar analysis-num-errors-found )



;; lemacs change: 4th elt can be column number.
(defconst analysis-error-regexp-alist
  '(
	;; Vantage custom Analyze script and vlint v1 format:
	;;test.vhdl, line 11: Error: msg
	;; vlint v2 simple format:
	;;test.vhdl, 11: Error: msg
	("\n\\([^,\n]+\\),[ \t]+\\(line[ \t]+\\)?\\([0-9]+\\):" 1 3)

	;; vlint v2 format with line and column number:
	;;test.vhdl, 11/15: Error: message text
	("\n\\([^,\n]+\\),[ \t]+\\([0-9]+\\)/\\([0-9]+\\):" 1 2 3)

	;; Leapfrog v2.1 error format
	;; duluth: *E,1000 (mux4.vhdl,35): message text
	("\n\\(duluth\\|bemidji\\):[ \t]+[*][EW],[0-9]+[ \t]+(\\([^,\n]+\\),\\([0-9]+\\)):" 2 3)

	;; Model Technology's vcom format:
	;; ERROR: foo.vhdl(219): message text
	("\n\\(ERROR\\|WARNING\\):[ \t]+\\([^(\n]+\\)(\\([0-9]+\\)):" 2 3)

	;; Synopsys VHDL analyzer vhdlan format:
	;;**Error: vhdlan,1081 foo.vhdl(48)
	("\n[*][*]\\(Error\\|Warning\\):[ \t]+vhdlan,[0-9]+[ \t]+\\([^(\n]+\\)(\\([0-9]+\\))" 2 3)

	;; Cadence NCSIM error format (without and with column number):
	;; ncvhdl_p: *E,EXPFOR (viper_tb.vhdl,558): message text
	("\nnc.*?:[ \t]+[*][EW],.*?(\\([^,\n]+\\),\\([0-9]+\\)):" 1 2)
	("\nnc.*?:[ \t]+[*][EW],.*?(\\([^,\n]+\\),\\([0-9]+\\)|\\([0-9]+\\)):" 1 2 3)
   )
  "Alist that specifies how to match errors in analyzer output.
   Each element has the form (REGEXP FILE-IDX LINE-IDX [COL-IDX]).
   If REGEXP matches, the FILE-IDX'th subexpression gives the file
   name, and the LINE-IDX'th subexpression gives the line number.
   COL-IDX'th subexpression gives the column number of the error.
   COL-IDX is optional, and if not present, beginning of line is used.")


;;;###autoload
(defvar analysis-search-path '( "." "./sources" )
  "*List of directories to search for source files named in error messages.
   Elements should be directory names, not file names of directories.
   nil as an element means to try the default directory.")



(defvar analyze-command "cv -m"
  "The shell command to run the VHDL analyzer.")

(defvar analyze-command-list nil
  "List of menu enties to start the VHDL analyzer. If this list is not nil
   the analyse-command variable is ignored and an submenu is created with
   all analyse commands in this list
   Example: (setq analyze-command-list '(\"cv -m\" \"cv -m -comp\"))")

(defvar analyze-check-command nil
  "The shell command to run the vlint VHDL Coding Standard checker.
   There is no default, as this depends on the vlint mode used.")


(defvar analyze-command-line (concat analyze-command " "
									 (and (buffer-file-name)
										  (file-name-nondirectory (buffer-file-name))))
  "Last shell command used to do a analysis; default for next analysis.
   Sometimes it is useful for files to supply local values for this variable.")


(defconst analysis-enter-directory-regexp
  ": Entering directory `\\(.*\\)'$"
  "Regular expression matching lines that indicate a new current directory.
   This must contain one \\(, \\) pair around the directory name.
   The default value matches lines printed by the `-w' option of GNU Make.")


(defconst analysis-leave-directory-regexp
  ": Leaving directory `\\(.*\\)'$"
  "Regular expression matching lines that indicate restoring current directory.
   This may contain one \\(, \\) pair around the name of the directory
   being moved from.  If it does not, the last directory entered \(by a
   line matching `analysis-enter-directory-regexp'\) is assumed.
   The default value matches lines printed by the `-w' option of GNU Make.")


(defvar analysis-directory-stack nil
  "Stack of previous directories for `analysis-leave-directory-regexp'.
   The head element is the directory the analysis was started in.")

;; History of analyze commands.
(defvar analyze-history nil)



;=============================================================================;
;                          defun analyze                                      ;
;=============================================================================;

;;;###autoload
(defun analyze (command &optional noquery)
  "Analyze the program including the current buffer.  Default: run `Analyze'.
   Runs COMMAND, a shell command, in a separate process asynchronously
   with output going to the buffer `*analysis*'.

   With optional argument NOQUERY non-nil, save all modified buffers without
   asking first.
   You can then use the command \\[analyze-next-error] to find the next
   error message and move to the source code that caused it.

   To run more than one analysis at once, start one and rename the
   \`*analysis*' buffer to some other name with \\[rename-buffer].
   Then start the next one.

   The name used for the buffer is actually whatever is returned by
   the function in `analysis-buffer-name-function', so you can set that
   to a function that generates a unique name."
  (interactive (list (read-string "Analyze command: "
								  (concat analyze-command " "
										  (file-name-nondirectory (buffer-file-name)))
								  'analyze-history)))
  (setq analyze-command-line command)
  (if noquery
	  (save-some-buffers t nil)
	(save-some-buffers nil nil))
  (analyze-internal analyze-command-line "No more errors"))


(defun analyze-with-command (analyze-cmd &optional noquery)
  (interactive)
  (setq analyze-command analyze-cmd)
  (setq analyze-command-line
		(read-string "Analyze command: "
					 (concat analyze-cmd " "
							 (file-name-nondirectory (buffer-file-name)))
					 'analyze-history))
  (if noquery
	  (save-some-buffers t nil)
	(save-some-buffers nil nil))
  (analyze-internal analyze-command-line "No more errors"))


(defun analyze-buffer-with-dialog (&optional noquery)
   (interactive)
   (let ((command (concat analyze-command " " (file-name-nondirectory (buffer-file-name)))))
      (if (y-or-n-p (concat command " ? "))
         (progn
            (setq analyze-history (append (list command) analyze-history))
            (analyze command noquery)
         )
         (message (concat command " cancelled ..."))
      )
   )
)
   

(defun analyze-noquery ()
  "Run analyze saving all modified buffers without querying."
  (interactive)
  (analyze analyze-command-line 'noquery))



;=============================================================================;
;                           defun analyze-check                               ;
;=============================================================================;

(defun analyze-check-beh (command &optional noquery)
  "Run vlint checker , with user-specified args, and collect output in a buffer.
While vlint runs asynchronously, you can use the \\[analyze-next-error] command
to find the code lines that vlint refer to."
  (interactive (list (read-string "Check command: "
								  (concat "vlint -b " (file-name-nondirectory (buffer-file-name)))
								  'analyze-history)))
  (setq analyze-check-command command)
  (if noquery
	  (save-some-buffers t nil)
	(save-some-buffers nil nil))
  (analyze-internal analyze-check-command "No more vlint messages"
					"VHDL Coding Standard Check"))

(defun analyze-check-beh-with-dialog(&optional noquery)
   (interactive)
   (let ((command (concat "vlint -b " (file-name-nondirectory (buffer-file-name)))))
      (if (y-or-n-p (concat command " ? "))
         (progn
            (setq analyze-history (append (list command) analyze-history))
            (analyze-check-beh command noquery)
         )
         (message (concat command " cancelled ..."))
      )
   )
)

(defun analyze-check-rtl (command &optional noquery)
  "Run vlint checker , with user-specified args, and collect output in a buffer.
While vlint runs asynchronously, you can use the \\[analyze-next-error] command
to find the code lines that vlint refer to."
  (interactive (list (read-string "Check command: "
								  (concat "vlint -r " (file-name-nondirectory (buffer-file-name)))
								  'analyze-history)))
  (setq analyze-check-command command)
  (if noquery
	  (save-some-buffers t nil)
	(save-some-buffers nil nil))
  (analyze-internal analyze-check-command "No more vlint messages"
					"VHDL Coding Standard Check"))

(defun analyze-check-rtl-with-dialog(&optional noquery)
   (interactive)
   (let ((command (concat "vlint -r " (file-name-nondirectory (buffer-file-name)))))
      (if (y-or-n-p (concat command " ? "))
         (progn
            (setq analyze-history (append (list command) analyze-history))
            (analyze-check-rtl command noquery)
         )
         (message (concat command " cancelled ..."))
      )
   )
)


(defun analyze-check-noquery ()
  "Run vhdl-analyze saving all modified buffers without querying."
  (interactive)
  (analyze-check-beh analyze-check-command 'noquery))



;=============================================================================;
;                         defun analyze-internal                              ;
;=============================================================================;

(defun analyze-internal (command error-message
				 &optional name-of-mode parser regexp-alist
				 name-function)
  "Run analysis command COMMAND (low level interface).
ERROR-MESSAGE is a string to print if the user asks to see another error
and there are no more errors.  Third argument NAME-OF-MODE is the name
to display as the major mode in the analysis buffer.

Fourth arg PARSER is the error parser function (nil means the default).  Fifth
arg REGEXP-ALIST is the error message regexp alist to use (nil means the
default).  Sixth arg NAME-FUNCTION is a function called to name the buffer (nil
means the default).  The defaults for these variables are the global values of
\`analysis-parse-errors-function', `analysis-error-regexp-alist', and
\`analysis-buffer-name-function', respectively.

Returns the analysis buffer created."
  (let (outbuf)
    (save-excursion
      (or name-of-mode
	  (setq name-of-mode "Analysis"))
      (setq outbuf
	    (get-buffer-create
	     (funcall (or name-function analysis-buffer-name-function
			  (function (lambda (mode)
				      (concat "*Analysis-Results*"))))
		      name-of-mode)))
      (set-buffer outbuf)
      (let ((comp-proc (get-buffer-process (current-buffer))))
	(if comp-proc
	    (if (or (not (eq (process-status comp-proc) 'run))
		    (yes-or-no-p
		     (format "A %s process is running; kill it? "
			     name-of-mode)))
		(condition-case ()
		    (progn
		      (interrupt-process comp-proc)
		      (sit-for 1)
		      (delete-process comp-proc))
		  (error nil))
	      (error "Cannot have two processes in `%s' at once"
		     (buffer-name))
	      )))
      ;; In case the analysis buffer is current, make sure we get the global
      ;; values of analysis-error-regexp-alist, etc.
      (kill-all-local-variables))
    (let ((regexp-alist (or regexp-alist analysis-error-regexp-alist))
	  (parser (or parser analysis-parse-errors-function))
	  (thisdir default-directory)
	  outwin)
      (save-excursion
	;; Clear out the analysis buffer unless analysis-append-results is t
    ;; and make it writable.
	;; Change its default-directory to the directory where the analysis
	;; will happen, and insert a `cd' command to indicate this.
		(set-buffer outbuf)
		(setq buffer-read-only nil)
		(erase-buffer)
		(setq default-directory thisdir)
		(insert "cd " thisdir "\n" command "\n")
		(set-buffer-modified-p nil))
      ;; If we're already in the analysis buffer, go to the end
      ;; of the buffer, so point will track the analysis output.
	  (if (eq outbuf (current-buffer))
		  (goto-char (point-max)))
	  ;; Pop up the analysis buffer.
	  (setq outwin (display-buffer outbuf))
	  (save-excursion
		(set-buffer outbuf)
		(analysis-mode)
		(buffer-disable-undo (current-buffer))
		;; (setq buffer-read-only t)  ;;; Non-ergonomic.
		(set (make-local-variable 'analysis-parse-errors-function) parser)
		(set (make-local-variable 'analysis-error-message) error-message)
		(set (make-local-variable 'analysis-error-regexp-alist) regexp-alist)
		(setq default-directory thisdir
			  analysis-directory-stack (list default-directory))
		(set-window-start outwin (point-min))
		(setq mode-name name-of-mode)
		(or (eq outwin (selected-window))
			(set-window-point outwin (point-min)))
		(and analysis-window-height
			 (= (window-width outwin) (frame-width))
			 (let ((w (selected-window)))
			   (unwind-protect
				   (progn
					 (select-window outwin)
					 (enlarge-window (- analysis-window-height
										(window-height))))
				 (select-window w))))
		;; Start the analysis.
		(let* (
			   ;;(shcmd (format "'cd %s; %s'" default-directory command))
			   ;; replaced command by shell command to ensure we are in the right
			   ;; working dir even if the user cd's in his .cshrc file or whatever
			   (process-connection-type nil) ; use pipe (not pty) for process start
			   (proc (start-process (downcase mode-name)
									outbuf "/bin/sh" "-c"
									(concat "cd " default-directory "; " command))))
			   ;;(proc (start-process-shell-command (downcase mode-name) outbuf shcmd)))
		  (set-process-sentinel proc 'analysis-sentinel)
		  (set-process-filter proc 'analysis-filter)
		  (set-marker (process-mark proc) (point) outbuf)
		  (setq analysis-in-progress (cons proc analysis-in-progress)))))
	;; Make it so the next C-x ` will use this buffer.
	(setq analysis-last-buffer outbuf)))


(defvar analysis-minor-mode-map
  (let ((map (make-sparse-keymap)))
    (set-keymap-name map 'analysis-minor-mode-map)
    (define-key map "\C-c\C-c" 'analyze-goto-error)
    (define-key map "\C-c\C-k" 'kill-analysis)
    (define-key map "\M-n" 'analysis-next-error)
    (define-key map "\M-p" 'analysis-previous-error)
    (define-key map "\M-{" 'analysis-previous-file)
    (define-key map "\M-}" 'analysis-next-file)
    map)
  "Keymap for `analysis-minor-mode'.")

(defvar analysis-mode-map
  (let ((map (make-sparse-keymap)))
    (set-keymap-parents map analysis-minor-mode-map)
    (set-keymap-name map 'analysis-mode-map)
    (define-key map " " 'scroll-up)
    (define-key map "\^?" 'scroll-down)
    (define-key map 'button2 'analyze-mouse-goto-error)
    (define-key map 'button3 'analyze-popup-menu)
    map)
  "Keymap for analysis log buffers.
`analysis-minor-mode-map' is a parent of this.")

(defun analysis-mode ()
  "Major mode for analysis log buffers.
\\<analysis-mode-map>To visit the source for a line-numbered error,
move point to the error message line and type \\[analyze-goto-error],
or click on the line with \\[analyze-mouse-goto-error].
There is a menu of commands on \\[analyze-popup-menu].
To kill the analysis, type \\[kill-analysis].

Runs `analysis-mode-hook' with `run-hooks' (which see)."
  (interactive)
  (fundamental-mode)
  (use-local-map analysis-mode-map)
  (setq major-mode 'analysis-mode
	mode-name "Analysis")
  (analysis-setup)
  (run-hooks 'analysis-mode-hook))

;; Prepare the buffer for the analysis parsing commands to work.
(defun analysis-setup ()
  ;; Make the buffer's mode line show process state.
  (setq modeline-process '(": %s"))
  (set (make-local-variable 'analysis-error-list) nil)
  (set (make-local-variable 'analysis-old-error-list) nil)
  (set (make-local-variable 'analysis-parsing-end) 1)
  (set (make-local-variable 'analysis-directory-stack) nil)
  (setq analysis-last-buffer (current-buffer))
  ;; Lemacs change: highlight lines, install menubar.
  (require 'mode-motion)
  (setq mode-motion-hook 'mode-motion-highlight-line)
  )

(defvar analysis-minor-mode nil
  "Non-nil when in analysis-minor-mode.
In this minor mode, all the error-parsing commands of the
Analysis major mode are available.")
(make-variable-buffer-local 'analysis-minor-mode)

(or (assq 'analysis-minor-mode minor-mode-alist)
    (setq minor-mode-alist (cons '(analysis-minor-mode " Analysis")
				 minor-mode-alist)))
;(or (assq 'analysis-minor-mode minor-mode-map-alist)
;    (setq minor-mode-map-alist (cons (cons 'analysis-minor-mode
;					   analysis-minor-mode-map)
;				     minor-mode-map-alist)))

;;;###autoload
(defun analysis-minor-mode (&optional arg)
  "Toggle analysis minor mode.
With arg, turn analysis mode on if and only if arg is positive.
See `analysis-mode'."
  (interactive "P")
  (if (setq analysis-minor-mode (if (null arg)
				       (null analysis-minor-mode)
				     (> (prefix-numeric-value arg) 0)))
      (analysis-setup)))

;; Called when analysis process changes state.
(defun analysis-sentinel (proc msg)
  "Sentinel for analysis buffers."
  (let ((buffer (process-buffer proc)))
    (if (memq (process-status proc) '(signal exit))
	(progn
	  (if (null (buffer-name buffer))
	      ;; buffer killed
	      (set-process-buffer proc nil)
	    (let ((obuf (current-buffer))
		  omax opoint)
	      ;; save-excursion isn't the right thing if
	      ;; process-buffer is current-buffer
	      (unwind-protect
		  (progn
		    ;; Write something in the analysis buffer
		    ;; and hack its mode line.
		    (set-buffer buffer)
		    (let ((buffer-read-only nil))
		      (setq omax (point-max)
			    opoint (point))
		      (goto-char omax)
		      ;; Record where we put the message, so we can ignore it
		      ;; later on.
		      (insert ?\n mode-name " " msg)
		      (forward-char -1)
		      (insert " at " (substring (current-time-string) 0 19))
		      (forward-char 1)
		      (setq modeline-process
			    (concat ": "
				    (symbol-name (process-status proc))))
		      ;; Since the buffer and mode line will show that the
		      ;; process is dead, we can delete it now.  Otherwise it
		      ;; will stay around until M-x list-processes.
		      (delete-process proc)
		      ;; Force mode line redisplay soon.
		      (set-buffer-modified-p (buffer-modified-p)))
		    (if (and opoint (< opoint omax))
			(goto-char opoint))
		    (if analysis-finish-function
			(funcall analysis-finish-function buffer msg)))
		(set-buffer obuf))))
	  (setq analysis-in-progress (delq proc analysis-in-progress))
	  ))))

(defun analysis-filter (proc string)
  "Process filter for analysis buffers.
Just inserts the text, but uses `insert-before-markers'."
  (save-excursion
    (set-buffer (process-buffer proc))
    (let ((buffer-read-only nil))
      (save-excursion
	(goto-char (process-mark proc))
	(insert-before-markers string)
	(set-marker (process-mark proc) (point))))))

;; Return the cdr of analysis-old-error-list for the error containing point.
(defun analyze-error-at-point ()
  (analyze-reinitialize-errors nil (point))
  (let ((errors analysis-old-error-list))
    (while (and errors
		(> (point) (car (car errors))))
      (setq errors (cdr errors)))
    errors))

(defun analysis-next-error (n)
  "Move point to the next error in the analysis buffer.
Does NOT find the source line like \\[analyze-next-error]."
  (interactive "p")
  (or (analysis-buffer-p (current-buffer))
      (error "Not in a analysis buffer."))
  (setq analysis-last-buffer (current-buffer))

  (let ((errors (analyze-error-at-point)))

    ;; Move to the error after the one containing point.
    (goto-char (car (if (< n 0)
			(let ((i 0)
			      (e analysis-old-error-list))
			  ;; See how many cdrs away ERRORS is from the start.
			  (while (not (eq e errors))
			    (setq i (1+ i)
				  e (cdr e)))
			  (if (> (- n) i)
			      (error "Moved back past first error")
			    (nth (+ i n) analysis-old-error-list)))
		      (let ((analysis-error-list (cdr errors)))
			(analyze-reinitialize-errors nil nil n)
			(if analysis-error-list
			    (nth (1- n) analysis-error-list)
			  (error "Moved past last error"))))))))

(defun analysis-previous-error (n)
  "Move point to the previous error in the analysis buffer.
Does NOT find the source line like \\[analyze-next-error]."
  (interactive "p")
  (analysis-next-error (- n)))


;; Given an elt of `analysis-error-list', return an object representing
;; the referenced file which is equal to (but not necessarily eq to) what
;; this function would return for another error in the same file.
(defsubst analysis-error-filedata (data)
  (setq data (cdr data))
  (if (markerp data)
      (marker-buffer data)
    (car data)))

;; Return a string describing a value from analysis-error-filedata.
;; This value is not necessarily useful as a file name, but should be
;; indicative to the user of what file's errors are being referred to.
(defsubst analysis-error-filedata-file-name (filedata)
  (if (bufferp filedata)
      (buffer-file-name filedata)
    (car filedata)))

(defun analysis-next-file (n)
  "Move point to the next error for a different file than the current one."
  (interactive "p")
  (or (analysis-buffer-p (current-buffer))
      (error "Not in a analysis buffer."))
  (setq analysis-last-buffer (current-buffer))

  (let ((reversed (< n 0))
	errors filedata)

    (if (not reversed)
	(setq errors (or (analyze-error-at-point)
			 (error "Moved past last error")))

      ;; Get a reversed list of the errors up through the one containing point.
      (analyze-reinitialize-errors nil (point))
      (setq errors (reverse analysis-old-error-list)
	    n (- n))

      ;; Ignore errors after point.  (car ERRORS) will be the error
      ;; containing point, (cadr ERRORS) the one before it.
      (while (and errors
		  (< (point) (car (car errors))))
	(setq errors (cdr errors))))

    (while (> n 0)
      (setq filedata (analysis-error-filedata (car errors)))

      ;; Skip past the following errors for this file.
      (while (equal filedata
		    (analysis-error-filedata
		     (car (or errors
			      (if reversed
				  (error "%s the first erring file"
					 (analysis-error-filedata-file-name
					  filedata))
				(let ((analysis-error-list nil))
				  ;; Parse some more.
				  (analyze-reinitialize-errors nil nil 2)
				  (setq errors analysis-error-list)))
			      (error "%s is the last erring file"
				     (analysis-error-filedata-file-name
				      filedata))))))
	(setq errors (cdr errors)))

      (setq n (1- n)))

    ;; Move to the following error.
    (goto-char (car (car (or errors
			     (if reversed
				 (error "This is the first erring file")
			       (let ((analysis-error-list nil))
				 ;; Parse the last one.
				 (analyze-reinitialize-errors nil nil 1)
				 analysis-error-list))))))))

(defun analysis-previous-file (n)
  "Move point to the previous error for a different file than the current one."
  (interactive "p")
  (analysis-next-file (- n)))


(defun kill-analysis ()
  "Kill the process made by the \\[analyze] command."
  (interactive)
  (let ((buffer (analysis-find-buffer)))
    (if (get-buffer-process buffer)
	(interrupt-process (get-buffer-process buffer))
      (error "The analysis process is not running."))))


;; Parse any new errors in the analysis buffer,
;; or reparse from the beginning if the user has asked for that.
(defun analyze-reinitialize-errors (argp &optional limit-search find-at-least)
  (save-excursion
    (set-buffer analysis-last-buffer)
    ;; If we are out of errors, or if user says "reparse",
    ;; discard the info we have, to force reparsing.
    (if (or (eq analysis-error-list t)
	    (consp argp))
	(progn (analysis-forget-errors)
	       (setq analysis-parsing-end 1)))
    (if (and analysis-error-list
	     (or (not limit-search)
		 (> analysis-parsing-end limit-search))
	     (or (not find-at-least)
		 (>= (length analysis-error-list) find-at-least)))
	;; Since analysis-error-list is non-nil, it points to a specific
	;; error the user wanted.  So don't move it around.
	nil

      ;; lemacs change: if the analysis buffer is already visible
      ;; in a window, use that instead of thrashing the display.
      (let ((w (get-buffer-window analysis-last-buffer)))
	(if w
	    (select-window w)
	  (switch-to-buffer analysis-last-buffer)))

      (set-buffer-modified-p nil)
      (if (< analysis-parsing-end (point-max))
	  (let ((at-start (= analysis-parsing-end 1)))
	    (funcall analysis-parse-errors-function
		     limit-search find-at-least)
	    ;; Remember the entire list for analysis-forget-errors.
	    ;; If this is an incremental parse, append to previous list.
	    (if at-start
		(setq analysis-old-error-list analysis-error-list)
	      (setq analysis-old-error-list
		    (nconc analysis-old-error-list analysis-error-list)))
	    )))))



(defun analyze-goto-error (&optional argp)
  "Visit the source for the error message point is on.
Use this command in a analysis log buffer.
\\[universal-argument] as a prefix arg means to reparse the buffer's error messages first;
other kinds of prefix arguments are ignored."
  (interactive "P")
  (or (analysis-buffer-p (current-buffer))
      (error "Not in a analysis buffer."))
  (setq analysis-last-buffer (current-buffer))
  (analyze-reinitialize-errors argp (point))

  ;; Move to bol; the marker for the error on this line will point there.
  (beginning-of-line)

  ;; Move analysis-error-list to the elt of analysis-old-error-list
  ;; we want.
  (setq analysis-error-list analysis-old-error-list)
  (while (and analysis-error-list
	      (> (point) (car (car analysis-error-list))))
    (setq analysis-error-list (cdr analysis-error-list)))

  ;; Move to another window, so that analyze-next-error's window changes
  ;; result in the desired setup.
  (or (one-window-p)
      (progn
	(other-window -1)
	;; other-window changed the selected buffer,
	;; but we didn't want to do that.
	(set-buffer analysis-last-buffer)))

  (analyze-next-error 1))



(defun analyze-mouse-goto-error (event)
  "Visit the source for the error under the mouse.
Use this command in a analysis log buffer."
  (interactive "e")
  (mouse-set-point event)
  (beginning-of-line)
  (analyze-goto-error))



(defun analysis-buffer-p (buffer)
  (assq 'analysis-error-list (buffer-local-variables buffer)))



;; Return a analysis buffer.
;; If the current buffer is a analysis buffer, return it.
;; If analysis-last-buffer is set to a live buffer, use that.
;; Otherwise, look for a analysis buffer and signal an error
;; if there are none.
(defun analysis-find-buffer (&optional other-buffer)
  (if (and (not other-buffer)
	   (analysis-buffer-p (current-buffer)))
      ;; The current buffer is a analysis buffer.
      (current-buffer)
    (if (and analysis-last-buffer (buffer-name analysis-last-buffer)
	     (or (not other-buffer) (not (eq analysis-last-buffer
					     (current-buffer)))))
	analysis-last-buffer
      (let ((buffers (buffer-list)))
	(while (and buffers (or (not (analysis-buffer-p (car buffers)))
				(and other-buffer
				     (eq (car buffers) (current-buffer)))))
	  (setq buffers (cdr buffers)))
	(if buffers
	    (car buffers)
	  (or (and other-buffer
		   (analysis-buffer-p (current-buffer))
		   ;; The current buffer is a analysis buffer.
		   (progn
		     (if other-buffer
			 (message "This is the only analysis buffer."))
		     (current-buffer)))
	      (error "No analysis started!")))))))



;;;###autoload
(defun analyze-next-error (&optional argp)
  "Visit next analysis error message and corresponding source code.
This operates on the output from the \\[analyze] command.
If all preparsed error messages have been processed,
the error message buffer is checked for new ones.

A prefix arg specifies how many error messages to move;
negative means move back to previous error messages.
Just C-u as a prefix means reparse the error message buffer
and start at the first error.

\\[analyze-next-error] normally applies to the most recent analysis started,
but as long as you are in the middle of parsing errors from one analysis
output buffer, you stay with that analysis output buffer.

Use \\[analyze-next-error] in a analysis output buffer to switch to
processing errors from that analysis.

See variables `analysis-parse-errors-function' and
\`analysis-error-regexp-alist' for customization ideas."
  (interactive "P")
  (setq analysis-last-buffer (analysis-find-buffer))
  (analyze-reinitialize-errors argp nil
			       ;; We want to pass a number here only if
			       ;; we got a numeric prefix arg, not just C-u.
			       (and (not (consp argp))
				    (if (< (prefix-numeric-value argp) 1)
					0
				      (1- (prefix-numeric-value argp)))))
  ;; Make ARGP nil if the prefix arg was just C-u,
  ;; since that means to reparse the errors, which the
  ;; analyze-reinitialize-errors call just did.
  ;; Now we are only interested in a numeric prefix arg.
  (if (consp argp)
      (setq argp nil))
  (let (next-errors next-error)
    (save-excursion
      (set-buffer analysis-last-buffer)
      ;; analysis-error-list points to the "current" error.
      (setq next-errors
	    (if (> (prefix-numeric-value argp) 0)
		(nthcdr (1- (prefix-numeric-value argp))
			analysis-error-list)
	      ;; Zero or negative arg; we need to move back in the list.
	      (let ((n (1- (prefix-numeric-value argp)))
		    (i 0)
		    (e analysis-old-error-list))
		;; See how many cdrs away the current error is from the start.
		(while (not (eq e analysis-error-list))
		  (setq i (1+ i)
			e (cdr e)))
		(if (> (- n) i)
		    (error "Moved back past first error")
		  (nthcdr (+ i n) analysis-old-error-list))))
	    next-error (car next-errors))
      (while
	  (progn
	    (if (null next-error)
;; lemacs change by Barry Warsaw.
;; Without this, if you a "no more errors" error, then you can't do
;; previous-error or goto-error until you kill the buffer.
;		(progn
;		  (if argp (if (> (prefix-numeric-value argp) 0)
;			       (error "Moved past last error")
;			     (error "Moved back past first error")))
;		  (analysis-forget-errors)
;		  (error (concat analysis-error-message
;				 (and (get-buffer-process (current-buffer))
;				      (eq (process-status
;					   (get-buffer-process
;					    (current-buffer)))
;					  'run)
;				      " yet"))))

		(if (> (prefix-numeric-value argp) 0)
		    (error "Moved past last error")
		  (error "Moved back past first error"))

	      (setq analysis-error-list (cdr next-errors))
	      (if (null (cdr next-error))
		  ;; This error is boring.  Go to the next.
		  t
		(or (markerp (cdr next-error))
		    ;; This error has a filename/lineno pair.
		    ;; Find the file and turn it into a marker.
		    (let* ((fileinfo (car (cdr next-error)))
			   (buffer (analysis-find-file (cdr fileinfo)
							  (car fileinfo)
							  (car next-error))))
		      (if (null buffer)
			  ;; We can't find this error's file.
			  ;; Remove all errors in the same file.
			  (progn
			    (setq next-errors analysis-old-error-list)
			    (while next-errors
			      (and (consp (cdr (car next-errors)))
				   (equal (car (cdr (car next-errors)))
					  fileinfo)
				   (progn
				     (set-marker (car (car next-errors)) nil)
				     (setcdr (car next-errors) nil)))
			      (setq next-errors (cdr next-errors)))
			    ;; Look for the next error.
			    t)
			;; We found the file.  Get a marker for this error.
			;; analysis-old-error-list is a buffer-local
			;; variable, so we must be careful to extract its value
			;; before switching to the source file buffer.
			(let ((errors analysis-old-error-list)
			      ;; lemacs change (columns)
			      (last-line (nth 1 (cdr next-error)))
			      (column (nth 2 (cdr next-error)))
			      )
			  (set-buffer buffer)
			  (save-excursion
			    (save-restriction
			      (widen)
			      (goto-line last-line)
			      (beginning-of-line)
			      ;; lemacs change (columns)
			      (and column
				   (forward-char (1- column)))
			      (setcdr next-error (point-marker))
			      ;; Make all the other error messages referring
			      ;; to the same file have markers into the buffer.
			      (while errors
				(and (consp (cdr (car errors)))
				     (equal (car (cdr (car errors))) fileinfo)
				     (let ((this (cdr (cdr (car errors))))
					   ;; lemacs change (columns)
					   (lines (- (nth 1 (cdr (car errors)))
						     last-line))
					   (col (nth 2 (cdr (car errors)))))
				       (if (eq selective-display t)
					   (if (< lines 0)
					       (re-search-backward "[\n\C-m]"
								   nil 'end
								   (- lines))
					     (re-search-forward "[\n\C-m]"
								nil 'end
								lines))
					 (forward-line lines))
				       ;; lemacs change (columns)
				       (setq last-line (car this))
				       (if col
					   (let ((eol (save-excursion
							(end-of-line)
							(point))))
					     (forward-char  ; don't pass eol.
					      (min (1- col) (- eol (point))))))
				       (setcdr (car errors) (point-marker))))
				(setq errors (cdr errors)))))))))
		;; If we didn't get a marker for this error,
		;; go on to the next one.
		(not (markerp (cdr next-error))))))
	(setq next-errors analysis-error-list
	      next-error (car next-errors))))

    ;; Skip over multiple error messages for the same source location,
    ;; so the next C-x ` won't go to an error in the same place.
    (while (and analysis-error-list
		(equal (cdr (car analysis-error-list)) (cdr next-error)))
      (setq analysis-error-list (cdr analysis-error-list)))

    ;; Lemacs change: If a new window has to be displayed, select the other
    ;; window to avoid swapping the position of the analysis error buffer.
    (and (get-buffer-window (marker-buffer (car next-error)))
	(progn
	  (select-window (get-buffer-window (marker-buffer (car next-error))))
	  (other-window -1)))

    ;; We now have a marker for the position of the error.
    (switch-to-buffer (marker-buffer (cdr next-error)))
    (goto-char (cdr next-error))
    ;; If narrowing got in the way of
    ;; going to the right place, widen.
    (or (= (point) (marker-position (cdr next-error)))
	(progn
	  (widen)
	  (goto-char (cdr next-error))))

    ;; Show analysis buffer in other window, scrolled to this error.
    (let* ((pop-up-windows t)
	   (w (display-buffer (marker-buffer (car next-error)))))
      (set-window-point w (car next-error))
      (set-window-start w (car next-error)))))



;; lemacs addition
;;;###autoload
(defun analyze-previous-error (&optional argp)
  "\\[analyze-next-error] backwards."
  (interactive "P")
  (analyze-next-error (cond ((null argp) -1)
		    ((numberp argp) (- argp))
		    (t argp))))



;;;###autoload
(define-key ctl-x-map "`" 'analyze-next-error)



;; Find a buffer for file FILENAME.
;; Search the directories in analysis-search-path.
;; A nil in analysis-search-path means to try the
;; current directory, which is passed in DIR.
;; If FILENAME is not found at all, ask the user where to find it.
;; Pop up the buffer containing MARKER and scroll to MARKER if we ask the user.
(defun analysis-find-file (filename dir marker)
  (let ((dirs analysis-search-path)
	result name)
    (while (and dirs (null result))
      (setq name (expand-file-name filename (or (car dirs) dir))
	    result (and (file-exists-p name)
			(find-file-noselect name))
	    dirs (cdr dirs)))
    (or result
	;; The file doesn't exist.
	;; Ask the user where to find it.
	;; If he hits C-g, then the next time he does
	;; analyze-next-error, he'll skip past it.
	(progn
	  (let* ((pop-up-windows t)
		 (w (display-buffer (marker-buffer marker))))
	    (set-window-point w marker)
	    (set-window-start w marker))
	  (setq name
		(expand-file-name
		 (read-file-name
		  (format "Find this error in: (default %s) "
			  filename) dir filename t)))
	  (if (file-directory-p name)
	      (setq name (concat (file-name-as-directory name) filename)))
	  (if (file-exists-p name)
	      (find-file-noselect name))))))



;; Set analysis-error-list to nil, and unchain the markers that point to the
;; error messages and their text, so that they no longer slow down gap motion.
;; This would happen anyway at the next garbage collection, but it is better to
;; do it right away.
(defun analysis-forget-errors ()
  (while analysis-old-error-list
    (let ((next-error (car analysis-old-error-list)))
      (set-marker (car next-error) nil)
      (if (markerp (cdr next-error))
	  (set-marker (cdr next-error) nil)))
    (setq analysis-old-error-list (cdr analysis-old-error-list)))
  (setq analysis-error-list nil
	analysis-directory-stack nil))



(defun count-regexp-groupings (regexp)
  "Return the number of \\( ... \\) groupings in REGEXP (a string)."
  (let ((groupings 0)
	(len (length regexp))
	(i 0)
	c)
    (while (< i len)
      (setq c (aref regexp i)
	    i (1+ i))
      (cond ((= c ?\[)
	     ;; Find the end of this [...].
	     (while (and (< i len)
			 (not (= (aref regexp i) ?\])))
	       (setq i (1+ i))))
	    ((= c ?\\)
	     (if (< i len)
		 (progn
		   (setq c (aref regexp i)
			 i (1+ i))
		   (if (= c ?\))
		       ;; We found the end of a grouping,
		       ;; so bump our counter.
		       (setq groupings (1+ groupings))))))))
    groupings))



(defun analysis-parse-errors (limit-search find-at-least)
  "Parse the current buffer as VHDL analyzer error messages.
   See variable `analysis-parse-errors-function' for the interface it uses."
  (setq analysis-error-list nil)
  (message "Parsing error messages...")
  (let (;;text-buffer
	orig orig-expanded parent-expanded
	regexp enter-group leave-group error-group
	alist subexpr error-regexp-groups
	(found-desired nil)
	(analysis-num-errors-found 0))

    ;; Don't reparse messages already seen at last parse.
    (goto-char analysis-parsing-end)
    ;; Don't parse the first two lines as error messages.
    ;; This matters for grep.
    (if (bobp)
	(progn
	  (forward-line 2)
	  ;; Move back so point is before the newline.
	  ;; This matters because some error regexps use \n instead of ^
	  ;; to be faster.
	  (forward-char -1)))

    ;; Analyze all the regexps we want to search for into one.
    (setq regexp (concat "\\(" analysis-enter-directory-regexp "\\)\\|"
			 "\\(" analysis-leave-directory-regexp "\\)\\|"
			 "\\(" (mapconcat (function
					   (lambda (elt)
					     (concat "\\(" (car elt) "\\)")))
					  analysis-error-regexp-alist
					  "\\|") "\\)"))

    ;; Find out how many \(...\) groupings are in each of the regexps, and set
    ;; *-GROUP to the grouping containing each constituent regexp (whose
    ;; subgroups will come immediately thereafter) of the big regexp we have
    ;; just constructed.
    (setq enter-group 1
	  leave-group (+ enter-group
			 (count-regexp-groupings
			  analysis-enter-directory-regexp)
			 1)
	  error-group (+ leave-group
			 (count-regexp-groupings
			  analysis-leave-directory-regexp)
			 1))

    ;; Analyze an alist (IDX FILE LINE), where IDX is the number of the
    ;; subexpression for an entire error-regexp, and FILE and LINE are the
    ;; numbers for the subexpressions giving the file name and line number.
    (setq alist (or analysis-error-regexp-alist
		    (error "analysis-error-regexp-alist is empty!"))
	  subexpr (1+ error-group))
    (while alist
      (setq error-regexp-groups (cons (list subexpr
					    (+ subexpr (nth 1 (car alist)))
					    ;; lemacs change (columns)
					    (+ subexpr (nth 2 (car alist)))
					    (let ((col (nth 3 (car alist))))
					      (and col
						   (+ subexpr col))))
				      error-regexp-groups))
      (setq subexpr (+ subexpr 1 (count-regexp-groupings (car (car alist)))))
      (setq alist (cdr alist)))

    (setq orig default-directory)
    (setq orig-expanded (file-truename orig))
    (setq parent-expanded (expand-file-name "../" orig-expanded))

    (while (and (not found-desired)
		;; We don't just pass LIMIT-SEARCH to re-search-forward
		;; because we want to find matches containing LIMIT-SEARCH
		;; but which extend past it.
		(re-search-forward regexp nil t))

      ;; Figure out which constituent regexp matched.
      (cond ((match-beginning enter-group)
	     ;; The match was the enter-directory regexp.
	     (let ((dir
		    (file-name-as-directory
		     (expand-file-name
		      (buffer-substring (match-beginning (+ enter-group 1))
					(match-end (+ enter-group 1)))))))
	       ;; The directory name in the "entering" message
	       ;; is a truename.  Try to convert it to a form
	       ;; like what the user typed in.
	       (setq dir
		     (analyze-abbreviate-directory dir orig orig-expanded
						   parent-expanded))
	       (setq analysis-directory-stack
		     (cons dir analysis-directory-stack))
	       (and (file-directory-p dir)
		    (setq default-directory dir))))

	    ((match-beginning leave-group)
	     ;; The match was the leave-directory regexp.
	     (let ((beg (match-beginning (+ leave-group 1)))
		   (stack analysis-directory-stack))
	       (if beg
		   (let ((dir
			  (file-name-as-directory
			   (expand-file-name
			    (buffer-substring beg
					      (match-end (+ leave-group
							    1)))))))
		     ;; The directory name in the "entering" message
		     ;; is a truename.  Try to convert it to a form
		     ;; like what the user typed in.
		     (setq dir
			   (analyze-abbreviate-directory dir orig orig-expanded
							 parent-expanded))
		     (while (and stack
				 (not (string-equal (car stack) dir)))
		       (setq stack (cdr stack)))))
	       (setq analysis-directory-stack (cdr stack))
	       (setq stack (car analysis-directory-stack))
	       (if stack
		   (setq default-directory stack))
	       ))

	    ((match-beginning error-group)
	     ;; The match was the composite error regexp.
	     ;; Find out which individual regexp matched.
	     (setq alist error-regexp-groups)
	     (while (and alist
			 (null (match-beginning (car (car alist)))))
	       (setq alist (cdr alist)))
	     (if alist
		 (setq alist (car alist))
	       (error "analysis-parse-errors: impossible regexp match!"))

	     ;; Extract the file name and line number from the error message.
	     (let ((beginning-of-match (match-beginning 0)) ;looking-at nukes
		   (filename
		    (cons default-directory
			  (buffer-substring (match-beginning (nth 1 alist))
					    (match-end (nth 1 alist)))))
		   (linenum (save-restriction
			      (narrow-to-region
			       (match-beginning (nth 2 alist))
			       (match-end (nth 2 alist)))
			      (goto-char (point-min))
			      ;; lemacs change (columns)
			      (save-match-data
				(if (looking-at "[0-9]")
				    (read (current-buffer))))
			      ))
		   ;; lemacs change (columns)
		   (column (and (nth 3 alist)
				(save-restriction
				  (narrow-to-region
				   (match-beginning (nth 3 alist))
				   (match-end (nth 3 alist)))
				  (goto-char (point-min))
				  (if (looking-at "[0-9]")
				      (read (current-buffer))))))
		   )
	       ;; Locate the erring file and line.
	       ;; Cons a new elt onto analysis-error-list,
	       ;; giving a marker for the current analysis buffer
	       ;; location, and the file and line number of the error.
	       (save-excursion
		 (beginning-of-line 1)
		 (let ((this (cons (point-marker)
				   ;; lemacs change (columns)
				   (list filename linenum column)
				   )))
		   ;; Don't add the same source line more than once.
		   (if (equal (cdr this) (cdr (car analysis-error-list)))
		       nil
		     (setq analysis-error-list
			   (cons this
				 analysis-error-list))
		     (setq analysis-num-errors-found
			   (1+ analysis-num-errors-found)))))
	       (and find-at-least (>= analysis-num-errors-found
				      find-at-least)
		    ;; We have found as many new errors as the user wants.
		    ;; We continue to parse until we have seen all
		    ;; the consecutive errors in the same file,
		    ;; so the error positions will be recorded as markers
		    ;; in this buffer that might change.
		    (cdr analysis-error-list) ; Must check at least two.
		    (not (equal (car (cdr (nth 0 analysis-error-list)))
				(car (cdr (nth 1 analysis-error-list)))))
		    (progn
		      ;; Discard the error just parsed, so that the next
		      ;; parsing run can get it and the following errors in
		      ;; the same file all at once.  If we didn't do this, we
		      ;; would have the same problem we are trying to avoid
		      ;; with the test above, just delayed until the next run!
		      (setq analysis-error-list
			    (cdr analysis-error-list))
		      (goto-char beginning-of-match)
		      (setq found-desired t)))
	       )
	     )
	    (t
	     (error "analysis-parse-errors: known groups didn't match!")))

      (message "Parsing error messages...%d (%d%% of buffer)"
	       analysis-num-errors-found
	       (/ (* 100 (point)) (point-max)))

      (and limit-search (>= (point) limit-search)
	   ;; The user wanted a specific error, and we're past it.
	   (setq found-desired t)))
    (setq analysis-parsing-end (if found-desired
				      (point)
				    ;; We have searched the whole buffer.
				    (point-max))))
  (setq analysis-error-list (nreverse analysis-error-list))
  (message "Parsing error messages...done"))


;; If directory DIR is a subdir of ORIG or of ORIG's parent,
;; return a relative name for it starting from ORIG or its parent.
;; ORIG-EXPANDED is an expanded version of ORIG.
;; PARENT-EXPANDED is an expanded version of ORIG's parent.
;; Those two args could be computed here, but we run faster by
;; having the caller compute them just once.
(defun analyze-abbreviate-directory (dir orig orig-expanded parent-expanded)
  (if (and (> (length dir) (length orig-expanded))
	   (string= orig-expanded
		    (substring dir 0 (length orig-expanded))))
      (setq dir
	    (concat orig
		    (substring dir (length orig-expanded)))))
  (if (and (> (length dir) (length parent-expanded))
	   (string= parent-expanded
		    (substring dir 0 (length parent-expanded))))
    (setq dir
	  (concat (file-name-directory
		   (directory-file-name orig))
		  (substring dir (length parent-expanded)))))
  dir)



;=============================================================================;
;                          Leapfrog Vhelp support                             ;
;=============================================================================;

(defvar analysis-leapfrog-error-regexp
  "^\\(cv\\|ev\\|duluth\\|bemidji\\):[ \t]+\\*[EW],\\([0-9]+\\)"
  "Regular expression describing Cadence Leapfrog analyzer messages, for which
   more detailed information can be obtained with the vhelp utility.
   Within the regexp the 1st subpattern denotes the tool name and the second
   subpattern denotes the message id.")


(defvar vhelp-mode-map nil
  "Keymap used in vhelp mode buffers.")

(if (not vhelp-mode-map)
    (setq vhelp-mode-map (make-sparse-keymap)))

(define-key vhelp-mode-map "x"       'kill-this-buffer)
(define-key vhelp-mode-map "q"       'bury-buffer)
(define-key vhelp-mode-map '(return) 'bury-buffer)
(define-key vhelp-mode-map " "       'scroll-up)
(define-key vhelp-mode-map "d"       'scroll-up)
(define-key vhelp-mode-map "u"       'scroll-down)



(defvar vhelp-mode-hooks nil
  "A list of functions to be called when entering vhelp mode.")



(defun vhelp-mode ()
  "Major mode for viewing Cadence Vhelp extended Leapfrog diagnostics."
  (interactive)
  (kill-all-local-variables)
  (use-local-map vhelp-mode-map)
  (setq major-mode 'vhelp-mode)
  (setq mode-name "Vhelp")
  (run-hooks 'vhelp-mode-hook))



(defun analysis-leapfrog-error-p ()
  "Test if point is on a line with Cadence Leapfrog error message."
  (save-excursion
	(beginning-of-line 1)
	(looking-at analysis-leapfrog-error-regexp)))



(defun analyze-show-vhelp ()
  (interactive)
  (let ((tool (buffer-substring (match-beginning 1) (match-end 1)))
		(id   (buffer-substring (match-beginning 2) (match-end 2)))
		(vhlp (get-buffer-create "*Leapfrog Extended Diagnostics*")))
		;;(cds  (or (getenv "CDS_VHDL")
		;;		  (error "Environment variable CDS_VHDL must be set to run vhelp."))))
	;; make vhlp buffer current
	(set-buffer vhlp)
	(vhelp-mode)
	;; make sure buffer can be written
	(setq buffer-read-only nil)
	(erase-buffer)
	(message "running vhelp subprocess ...")
	(call-process "vhelp" nil vhlp t "-nocopyright" tool id)
	(set-buffer-modified-p nil)
	(setq buffer-read-only t)
	(message "running vhelp subprocess ... done")
	;; eventually bring window to front
	(switch-to-buffer-other-window vhlp)))



;=============================================================================;
;                                 XEmacs menus                                ;
;=============================================================================;


(defun analysis-errors-exist-p (&optional buffer)
  "Whether we are in a state where the `analyze-next-error' command will work,
   that is, whether there exist (or may exist) error targets in the
   *analyze* buffers."
  (or buffer
      (setq buffer (condition-case nil
			  (analysis-find-buffer)
			(error nil))))
  (and buffer
       (analysis-buffer-p buffer)
       (save-excursion
	 (set-buffer buffer)
	 ;; Has errors on the list, or needs to be parsed.
	 ;; But don't parse it now!
	 (or (not (null analysis-error-list))
	     (< analysis-parsing-end (point-max))))))



(defvar analysis-mode-menu
  '("Analysis Mode Commands"
    ["Analyze..."	        analyze                 t]
    ["Kill Analysis"	    kill-analysis           (get-buffer-process (current-buffer))]
    ["Leapfrog VHelp ..."	analyze-show-vhelp	    (analysis-leapfrog-error-p)]
    ["Goto Error"	        analyze-goto-error	    (analysis-errors-exist-p)]
    ["Next Error" 	        analyze-next-error	    (analysis-errors-exist-p)]
    ["Previous Error"	    analyze-previous-error	(analysis-errors-exist-p)]))



(defun analyze-popup-menu (e)
  "Pops the analyze menu up."
  (interactive "@e")
  (let* ((submenu (mapcar #'(lambda (string)
							  (vector string (list 'analyze string) t))
						  analyze-history))
		 (menu analysis-mode-menu)
		 (name "Analyze History")
		 (existing (assoc name menu)))
    (if existing
		(setcdr existing submenu)
      (nconc menu (list (cons name submenu))))
    (popup-menu menu)))


(provide 'analyze)

;;; analyze.el ends here
