;;============================================================================;;
;;                                                                            ;;
;;  File          :  vexpand.el                                               ;;
;;  Related Files :                                                           ;;
;;                                                                            ;;
;;  Author(s)     :  Werner Hack, Claus Mayer                                 ;;
;;  Organization  :  Lucent Technologies Nuremberg                            ;;
;;                                                                            ;;
;;  Creation Date :  21.05.1993                                               ;;
;;                                                                            ;;
;;  Usage         :  (require 'vexpand)                                       ;;
;;                   (vexpand "filename")                                     ;;
;;                                                                            ;;
;;  Description   :  Insert file FILENAME at point thereby replacing all      ;;
;;                   variable references of the form  %variable:width% with   ;;
;;                   the appropriate values.                                  ;;
;;                                                                            ;;
;;  Tested with   :  SunOs 4.1.3  lucid 19.6                                  ;;
;;                                                                            ;;
;;============================================================================;;
;;                               Copyright                                    ;;
;;============================================================================;;
;;                                                                            ;;
;;              Copyright (C) 1993-1999 by Lucent Technologies                ;;
;;                                                                            ;;
;;     All rights reserved. Reproduction in whole or part is prohibited       ;;
;;          without the written permission of the copyright owner.            ;;
;;                                                                            ;;
;;============================================================================;;

(provide 'vexpand)

(defconst vexpand-rcsid "$Id: vexpand.el,v 1.5 1998/12/15 13:13:38 ic_cma Exp $")

(defvar vexpand-insert-alist '(("[Mm]akefile$"       . "makefile.tmpl")
							   ("\\.mk$"             . "makefile.tmpl")
							   ("[Dd]istfile$"       . "distfile.tmpl")
							   ("\\.c$"              . "c.tmpl")
							   ("\\.cc$"             . "c++.tmpl")
							   ("\\.cpp$"            . "c++.tmpl")
							   ("\\.C$"              . "c++.tmpl")
							   ("\\.l$"              . "lex.tmpl")
							   ("\\.y$"              . "yacc.tmpl")
							   ("\\.ll$"             . "lade-lexer.tmpl")
							   ("\\.lg$"             . "lade-parser.tmpl")
							   ("\\.lt$"             . "lade-translator.tmpl")
							   ("\\.vhd[l]?$"        . "vhdl.tmpl")
							   ("\\.v$"              . "verilog.tmpl")
							   ("\\.dc$"             . "synopsys.tmpl")
							   ("\\.h$"              . "h.tmpl")
							   ("\\.el$"             . "elisp.tmpl")
							   ("\\.pl$"             . "perl.tmpl")
							   ("\\.memo$"           . "memo.tmpl")
							   ("\\.texi\\(nfo\\)?$" . "texinfo.tmpl")
							   ("\\.man$"            . "manpage.tmpl")
							   ("\\.[1-8]$"          . "manpage.tmpl"))
  "*Alist specifying text to insert by default into a newly created file.
   Elements look like (REGEXP . FILENAME). If the new file's name matches
   REGEXP, then the file FILENAME will be inserted into the buffer.
   Only the first matching element will be considered.")



(defvar vexpand-search-path '("."
                              "~/templates"
							  "~/lib/templates"
							  "/usr/lib/templates"
							  "/usr/local/lib/templates"
							  (concat (getenv "CADENV_HOME") "/.caddata/vhdl_devsys/misc/xemacs/templates")
							  (concat (getenv "VHDL_DEVSYS_DIR") "/misc/xemacs/templates"))
  "*List of directories to search for files to be included as specified
   in vexpand-insert-alist.")



(defconst vexpand-month-names
      '("Dec" "Nov" "Oct" "Sep" "Aug" "Jul"
	"Jun" "May" "Apr" "Mar" "Feb" "Jan")
      "Backward list of months as they appear in \\{current-time-string}")


(defun vexpand-month-number (abbr)
  "Return number of month 1-12."
  (catch 'found
    (let ((list vexpand-month-names))
      (while list
	(if (string= abbr (car list))
	    (throw 'found (length list))
	  (setq list (cdr list))))
      0)))


(defun vexpand-make-date-string ()
  (let* ((str (current-time-string))
	(day (format "%02d" (string-to-int (substring str 8 10))))
	(mon (format "%02d" (vexpand-month-number (substring str 4 7))))
	(year (substring str -4 nil)))
	(concat day "." mon "." year)))


(defun vexpand-file-base-name (name)
  "Returns the basename component of a pathname."
  (interactive)
  (let ((fname (file-name-nondirectory name)))
	(substring fname 0 (string-match "\\.[^.]*$" fname))))


(defun vexpand-file-extension (name)
  "Returns the extension of a filename."
  (interactive)
  (let* ((fname (file-name-nondirectory name))
		 (pos (string-match "\\.[^.]*$" fname)))
	(if pos
		(substring fname (1+ pos) nil)
	  "")))


(defun vexpand-in-alist (alist var)
  "Check in VAR is a car of alist and return the corresponding cdr."
  (interactive)
  (let ((val nil)
		(list alist))
	(while (and (not val) list)
	  (if (string= (car (car list)) var)
		  (setq val (cdr (car list)))
		(setq list (cdr list))))
	(setq val val)))



(defun vexpand (filename &rest var-alist)
  "Insert file FILENAME at point, replacing all variable references
   of the form %variable[:width]% with the appropriate values given by the
   association list VAR-ALIST ."
  (interactive "fFilename: ")
  (let* ((path-name (or (buffer-file-name) ""))
		 (dir-name  (or (file-name-directory path-name) ""))
		 (file-name (file-name-nondirectory path-name))
		 (base-name (vexpand-file-base-name path-name))
		 (file-ext  (vexpand-file-extension path-name))
		 (var) (VAR) (tmp) (val) (wid) (gap)
		 (__start__ nil))

	(insert-file-contents filename)

	(while (re-search-forward "%\\([-a-zA-Z0-9_]+\\)\\(:[0-9]+\\)?%"
							  (point-max) t)
	  (setq var (buffer-substring (match-beginning 1) (match-end 1)))
	  (setq VAR (upcase var))

	  ;; how can var be substituted ?
	  (cond
	   ;; ====== predefined variables ========
	   ((string= var "date")
		(setq val (vexpand-make-date-string)))
	   ((string= var "time")
		(setq val (substring (current-time-string) 11 19)))
	   ((string= var "pathname")
		(setq val path-name))
	   ((string= var "dirname")
		(setq val dir-name))
	   ((string= var "filename")
		(setq val file-name))
	   ((string= var "FILENAME")
		(setq val (upcase file-name)))
	   ((string= var "basename")
		(setq val base-name))
	   ((string= var "BASENAME")
		(setq val (upcase base-name)))
	   ((string= var "extension")
		(setq val file-ext))
	   ((string= var "EXTENSION")
		(setq val (upcase file-ext)))
	   ((string= var "username")
		(setq val (user-login-name)))
	   ((string= var "fullname")
		(setq val (user-full-name)))
	   ((string= var "hostname")
		(setq val (system-name)))
	   ((string= var "__start__")
		(if (not __start__)
			(setq __start__ (point-marker)))
		(setq val ""))
	   ;; ====== environment variables ========
	   ((getenv VAR)
		(setq val (getenv VAR)))
	   ;; ====== association list ========
	   ((setq tmp (vexpand-in-alist var-alist var))
		(setq val tmp))
	   ;; ====== default replacement: "" ========
	   (t
		(setq val "")))

	  (cond
	   ((match-beginning 2)		; optional width specified ?
		(setq wid (string-to-int (buffer-substring
								  (+ (match-beginning 2) 1) (match-end 2))))

		(if (<= wid (length val))
			;; truncate val to wid
			(setq val (substring val 0 wid))
		  ;; else pad string to specified width
		  (setq gap (- wid (length val)))
		  (setq val (concat val (make-string gap ?  ))))))

	  (replace-match val))
	(if __start__
		(progn
		  (goto-char __start__)
		  (set-marker __start__ nil)))))



(defun vexpand-insert-file ()
  "Insert default template file into a newly created file.
   Template file is determined by matching visited filename against entries
   in `vexpand-insert-alist'."
  (let ((alist vexpand-insert-alist)
		(name (buffer-file-name))
		(old-state (buffer-modified-p))
		(insert-file nil))

	;; find first matching alist entry:
	(while (and (not insert-file) alist)
	  (if (string-match (car (car alist)) name)
		  (setq insert-file (cdr (car alist)))
		(setq alist (cdr alist))))

	;; if such a file exists in vexpand-search-path, insert it:
	(if insert-file
		(let* ((dirs vexpand-search-path)
			   (file (concat (car dirs) "/" insert-file)))
		  (while (and (not (file-readable-p file)) dirs)
			(setq dirs (cdr dirs))
			(setq file (concat (car dirs) "/" insert-file)))
		  (if dirs
			  (progn
				(vexpand file)
				(set-buffer-modified-p old-state)))))))


