;;============================================================================;;
;;                                                                            ;;
;;   File          :  vhdl.el                                                 ;;
;;   Requires      :  analyze.el font-lock.el func-menu.el vexpand.el         ;;
;;                                                                            ;;
;;   Author(s)     :  Werner Hack, Martin Zeh, Claus Mayer                    ;;
;;   Organization  :  Lucent Technologies Nuremberg                           ;;
;;                                                                            ;;
;;   Creation Date :  22.06.1993                                              ;;
;;                                                                            ;;
;;   Usage         :  (load-library "vhdl")                                   ;;
;;                    (autoload 'vhdl-mode "vhdl")                            ;;
;;                                                                            ;;
;;   Description   :  Major mode for editing VHDL source files with           ;;
;;                    hooked abbreviations of all imortant VHDL constructs.   ;;
;;                                                                            ;;
;;   Tested with   :  SunOS 4.1.3  HP-UX 09.01 09.05 10.01 - 10.30            ;;
;;                    Solaris 2.3 - 2.5                                       ;;
;;                    xemacs v19.13 - v20.4                                   ;;
;;                                                                            ;;
;;============================================================================;;
;;                                Copyright                                   ;;
;;============================================================================;;
;;                                                                            ;;
;;            Copyright (C) 1993-1999 by Lucent Technologies                  ;;
;;                                                                            ;;
;;      All rights reserved. Reproduction in whole or part is prohibited      ;;
;;           without the written permission of the copyright owner.           ;;
;;                                                                            ;;
;;============================================================================;;


(defconst vhdl-rcsid "$Id: vhdl.el,v 2.13 1999/09/24 13:49:24 ic_cma Exp $")

(defconst vhdl-rcs-revision "$Revision: 2.13 $")
(defvar vhdl-revision "1.1")

(defconst vhdl-rcs-date "$Date: 1999/09/24 13:49:24 $")
(defvar vhdl-date "1970/01/01")

(if (string-match "Revision: \\([0-9.]+\\)" vhdl-rcs-revision)
    (setq vhdl-revision (substring vhdl-rcs-revision (match-beginning 1)
                                   (match-end 1))))

(if (string-match "Date: \\([0-9/]+\\)" vhdl-rcs-date)
    (setq vhdl-date (substring vhdl-rcs-date (match-beginning 1)
                                   (match-end 1))))

(defconst vhdl-version (concat vhdl-revision " " vhdl-date))

(defun vhdl-show-version ()
  "Display VHDL-mode version number in minibuffer."
  (interactive)
  (message "Using VHDL-mode v%s (C) Lucent Technologies Nuernberg"
           vhdl-version))

(setenv "TERM" "dumb") ; to avoid xterm-specific shell-output

;;============================================================================;;
;;                         user customizable variables                        ;;
;;============================================================================;;

(defvar vhdl-inhibit-toolbar nil
  "*Non-nil means don't use the specialized Info toolbar.")
  
(defvar vhdl-frame-block-comments nil
  "*Non-nil means: block comments must have trailing '--' at each end-of-line")

(defvar vhdl-indent 4
  "*The number of columns to indent syntactic levels in VHDL-Mode")


(defvar vhdl-start-comment-column 40
  "*Column to which trailing comments will be aligned to.")

(defvar vhdl-end-comment-column 80
  "*Column at which long comment lines will be wrapped.")

(defvar vhdl-comment-region-prefix "--+"
  "*A VHDL comment string used to indicate a temporary comment.")


(defvar vhdl-keyword-case-defun 'upcase
  "*Controls the character case of inserted VHDL keywords.
    Suitable functions include `upcase', `downcase' and `capitalize'
    and 'identity.")

(defvar vhdl-identifier-case-defun 'downcase
  "*Controls the insertion of newlines before and after certain colons.
    Suitable function  include `upcase', `downcase' and `capitalize'
    and 'identity.")

(defvar vhdl-attribute-case-defun 'upcase
  "*Controls the insertion of newlines before and after certain colons.
    Suitable functions include `upcase', `downcase' and `capitalize'
    and 'identity.")

(defvar vhdl-abbrev-mode t
  "switch for abbreviation abbrev-mode is on if vhdl-abbrev-mode is non nil")

(defvar vhdl-configuration-suffix "_cfg"
  "Suffix to generate the configuration name <entity>vhdl-configuration-suffix
   This variable is used bye insert default configuration")

(defvar vhdl-formatting-command "vfmt"
  "*The name of an external filter program to format a VHDL buffer.")

(defvar vhdl-format-region-arg "-k"
  "*Extra argument to the formatter to format regions")

(defvar vhdl-check-command "vlint"
  "*The name of an external filter program to check a VHDL buffer.")

(defvar vhdl-print-command "vpr"
  "*The name of an external program to print out a VHDL buffer.")


(defvar vhdl-auto-enclose-expr nil
  "*If non-nil means that a pair of parenthesis is placed around expressions
   in certain control constructs.")

(defvar vhdl-project-id-list '("no menu")
  "*List of VDS_PROJECTID's to choose via VHDL-menu
    example:
      (setq vhdl-project-id-list '(\"asic1\" \"asic2\"))")

(defvar vhdl-library-use-list '(("IEEE"
                                 "ieee"
                                 "ieee.std_logic_1164")
                                ("IEEE+UTIL"
                                 "ieee, util"
                                 "ieee.std_logic_1164"
                                 "util.std_logic_arith"
                                 "util.int_logic_synt"))
  "This list defines the submenu 'LIBRARY/USE' in the 'VHDL' menu
Syntax:  '( ( <menustring> <library> <use> ... ) ... )
The list can be modified with for example:
  (setq vhdl-library-use-list
      (append '((\"LIB1\"      \"lib1\"       \"lib1.pckg1\"
                                          \"lib1.pckg2\")
                (\"LIB2\"      \"lib2\"       \"lib2.pckg3\")
                (\"LIB1+LIB2\" \"lib1, lib2\" \"lib1.pckg1\"
                                          \"lib1.pckg2\"
                                          \"lib2.pckg3\"))
              vhdl-library-use-list))")

;;============================================================================;;
;;                                 constants                                  ;;
;;============================================================================;;

(defconst vhdl-simple-keyword-list
  '( "abs" "access" "after" "alias" "all" "and" "array" "assert" "attribute"
     "begin" "body" "buffer" "bus" "constant" "disconnect" "downto" "exit"
     "file" "for" "generate" "generic" "group" "guarded" "impure" "in" "inout"
     "inertial" "is" "label" "library" "linkage" "literal" "map" "mod" "nand"
	 "new" "next" "nor" "not" "null" "of" "on" "open" "or" "others" "out" 
	 "package" "port" "postponed" "pure" "range" "register" "reject" "rem"
     "report" "return" "rol" "ror" "select" "signal" "shared" "sla" "sll"
     "subtype" "sra" "srl" "then" "to" "transport" "type" "unaffected" "until"
	 "use" "variable" "wait" "with" "xnor" "xor" )
  "A list of VHDL keywords which are simply inserted into buffer
   with the desired keyword case. No expansion into structures
   takes place.")



(defconst vhdl-complex-keyword-list
  '( "architecture" "block" "case" "component" "configuration" "else" "elsif"
     "end" "entity" "function" "if" "loop" "procedure" "process" "record"
     "severity" "units" "when" "while" )
  "A list of VHDL keywords which are expanded into more complex structures.")


(defconst vhdl-keyword-list
  (append vhdl-simple-keyword-list vhdl-complex-keyword-list nil)
  "A list of all VHDL keywords.")


(defconst vhdl-keyword-regexp
  (concat "\\<\\(" (mapconcat 'identity vhdl-keyword-list "\\|") "\\)\\>")
  "Regular expression describing VHDL keywords.")


(defconst vhdl-short-keywords
  "\\<\\([Ii][Ff]\\|[Ii][Nn]\\|[Ii][Ss]\\|[Oo][Ff]\\|[Oo][Nn]\\|[Oo][Rr]\\|[Tt][Oo]\\)\\>"
  "Regexpr describing all two-letter VHDL keywords")


(defconst vhdl-attribute-list
  '( "active" "ascending" "base" "behavior" "delayed" "dirving" "driving_value" "event"
	 "foreign" "high" "image" "instance_name" "last_active" "last_event"
     "last_value" "left" "leftof" "length" "low" "path_name" "pos" "pred" "quiet" "range"
     "reverse_range" "right" "rightof" "simle_name" "succ" "stable" "structure" 
     "transaction" "val" "value")
  "List of strings  describing VHDL predefined attribute names.")


(defconst vhdl-attribute-regexp
  (concat "'\\s-*\\(" (mapconcat 'identity vhdl-attribute-list "\\|") "\\)\\>")
  "Regular expression describing VHDL predefined attribute names.")


;;============================================================================;;
;;                               internal variables                           ;;
;;============================================================================;;

(defvar vhdl-saved-entity-name ""
  "Name of the last entity offered as default in subsequent architecture
   statements.")


(defvar vhdl-saved-architecture-name ""
  "Name of the last architecture offered as default in subsequent configuration
   statements.")


(defvar vhdl-identifier-case-buffer-defun 'downcase-word
  "Controls the insertion of newlines before and after certain colons.
   Suitable function  include `upcase', `downcase' and `capitalize'
   and 'identity.")


;;============================================================================;;
;;                               font-lock support                            ;;
;;============================================================================;;

(require 'font-lock)

;; create two new faces 'font-lock-attribute-face and 'font-lock-error-message-face
;; which are not part of the original font-lock package

;; face for predefined attribute names
(make-face 'font-lock-attribute-face)

;; face used for error messages in Vantage interactive error correction buffers
(make-face 'font-lock-error-message-face)

(set-face-foreground  'font-lock-error-message-face "white")
(set-face-background  'font-lock-error-message-face "red")
(set-face-underline-p 'font-lock-error-message-face nil)


;; highlight attributes *before* keywords in order to get 'RANGE in
;; font-lock-attribute-face not in font-lock-keyword-face !!

(defconst vhdl-fontlock-keywords
  (list
   (cons "^\\*\\*\\*?\\(Error\\|Warning\\).*$"   'font-lock-error-message-face)
   (list vhdl-attribute-regexp                 1 'font-lock-attribute-face)
   (cons vhdl-keyword-regexp                     'font-lock-keyword-face))
  "Expressions to highlight in VHDL-mode.")


;;============================================================================;;
;;                            function-menu support                           ;;
;;============================================================================;;

(require 'func-menu)

(defconst vhdl-fume-function-name-regexp
  "^\\s-*\\(\\(\\sw+\\)\\s-*:\\s-*PROCESS\\)\\|\\(\\(ENTITY\\|ARCHITECTURE\\|CONFIGURATION\\|PACKAGE\\(\\s-+BODY\\)?\\|FUNCTION\\|PROCEDURE\\|TYPE\\|SUBTYPE\\|ALIAS\\|SIGNAL\\|CONSTANT\\|VARIABLE\\|COMPONENT\\)\\s-+\\(\\sw+\\)\\)"
  "Expression to get VHDL object names")

(setq fume-function-name-regexp-alist (cons '(vhdl-mode . vhdl-fume-function-name-regexp)
                                            fume-function-name-regexp-alist))

(defun vhdl-find-next-fume-object (buffer)
  "Searches for the next VHDL object in BUFFER."
  (set-buffer buffer)
  (let ((nobreak t)
        (found nil)
        matches
        identifier
        keyword
        position)
    (while nobreak
      (if (re-search-forward fume-function-name-regexp nil t)
          (progn
            (setq matches (match-data))
            (if (not (or (vhdl-in-comment-p) (vhdl-in-quote-p)))
                (progn
                  (store-match-data matches)
                  (setq nobreak nil)
                  (setq found t)
                  ;; for VHDL we have to make a distinction between PROCESSES and other
                  ;; statements because PROCESSES are preceded by their label !
                  (if (save-excursion
                        (goto-char (match-beginning 0))
                        (looking-at "\\s-*\\(\\sw+\\)\\s-*:\\s-*PROCESS"))
                      (progn
                        (setq position (match-beginning 1))
                        (setq keyword "PROCESS")
                        (setq identifier (buffer-substring (match-beginning 1) (match-end 1))))
                    ;; not a PROCESS statement
                    (store-match-data matches)
                    (setq position (match-beginning 6))
                    (setq keyword    (buffer-substring (match-beginning 4) (match-end 4)))
                    (setq identifier (buffer-substring (match-beginning 6) (match-end 6)))))))
        (setq nobreak nil)))
    (if found
        (cons (concat keyword " " identifier) position))))

(setq fume-find-function-name-method-alist (cons '(vhdl-mode . vhdl-find-next-fume-object)
                                                 fume-find-function-name-method-alist))

;;============================================================================;;
;;                               VHDL menus                                   ;;
;;============================================================================;;

(defvar vhdl-header-dir (or (getenv "HEADERDIR") (concat (getenv "CADENV_HOME") "/.caddata/vhdl_devsys/misc/hdr"))
  "The name of a directory where header files are expected for 'Include Header'.")

(require 'analyze)

(defun vhdl-get-analyze-menu (items)
  (interactive)
  (let ((ana-list (reverse analyze-command-list))
        (analyze-menu))
    (if (not ana-list)
        (setq ana-list (list analyze-command)))
    (while ana-list
      (if (not (equal analyze-command (car ana-list)))
          (setq analyze-menu (cons
                              (vector (car ana-list) (list 'analyze-with-command (car ana-list)) t)
                              analyze-menu)))
      (setq ana-list (cdr ana-list)))
    (setq analyze-menu (cons "-----" analyze-menu))
    (setq analyze-menu (cons
                        (vector analyze-command (list 'analyze-with-command analyze-command) t)
                        analyze-menu))
    analyze-menu))

(defun vhdl-insert-library-use (menu-str)
  "Insert the LIBRARY and USE statements defined by the
variable vhdl-library-use-list"
  (let ((lu-list vhdl-library-use-list)
        lu-sublist)
    (while (car lu-list)
      (if (equal menu-str (car (car lu-list)))
          (progn
            (setq lu-sublist (car lu-list))
            (vhdl-insert-keyword "library ")
            (vhdl-insert-identifier (concat (car (cdr lu-sublist)) ";\n"))
            (setq lu-sublist (cdr (cdr lu-sublist)))
            (while (car lu-sublist)
              (vhdl-insert-keyword "use ")
              (vhdl-insert-identifier (concat (car lu-sublist) "."))
              (vhdl-insert-keyword "all;\n")
              (setq lu-sublist (cdr lu-sublist)))))
      (setq lu-list (cdr lu-list)))))

(defun vhdl-insert-library-use-menu (items)
  "Filter function for submenu 'LIBRARY/USE' in the 'VHDL' menu.
The variable vhdl-library-use-list is used to build the menu"
  (let ((lu-list (reverse vhdl-library-use-list))
        menu-str menu)
    (while (car lu-list)
      (setq menu-str (car (car lu-list)))
      (setq menu (cons (vector menu-str (list 'vhdl-insert-library-use menu-str) t) menu))
      (setq lu-list (cdr lu-list)))
    menu))

(defun vhdl-get-prj-id-menu (items)
  "Filter function for submenu 'Include HEADER->Set Project Id'
in th 'VHDL' menu.
The variable vhdl-project-id-list is used to generate the menu items"
  (let ((list (reverse vhdl-project-id-list))
        (prjid (getenv "VDS_PROJECTID"))
        (menu '()))
    (while (car list)
      (if (not (equal prjid (car list)))
          (setq menu (cons (vector (car list) (list 'vhdl-set-prj-id (car list)) t) menu)))
      (setq list (cdr list)))
    (if prjid
        (progn
          (setq menu (cons "-----" menu))
          (setq menu (cons (vector prjid (list 'vhdl-set-prj-id prjid) t) menu))))
    menu))

(defun vhdl-set-prj-id (prjid)
  (interactive)
  (if (equal "no menu" prjid)
      (error "No variable vhdl-project-id-list defined"))
  (setenv "VDS_PROJECTID" prjid))

(defvar vhdl-pulldown-menu
  '("VHDL"
    ("Include Header"
     ["VHDL File Header"    (vexpand (concat vhdl-header-dir "/FileHeader"))    t]
     "-----"
     ["Entity"              (vexpand (concat vhdl-header-dir "/EntityHeader"))  t]
     ["Architecture"        (vexpand (concat vhdl-header-dir "/ArchHeader"))    t]
     ["Configuration"       (vexpand (concat vhdl-header-dir "/ConfigHeader"))  t]
     ["Package"             (vexpand (concat vhdl-header-dir "/PackageHeader")) t]
     ["Process"             (vexpand (concat vhdl-header-dir "/ProcessHeader")) t]
     ["Subprogram"          (vexpand (concat vhdl-header-dir "/SubProgHeader")) t]
     "-----"
     ("Set Project Id" :filter vhdl-get-prj-id-menu))
    ("Insert Default"
     ["COMPONENT"           (vhdl-insert-default-component)         t]
     ["PORT MAP"            (vhdl-insert-default-port-map)          t]
     ["SIGNAL LIST"         (vhdl-insert-default-signals)           t]
     ["CONFIGURATION"       (vhdl-insert-default-configuration)     t])
    ("Insert LIBRARY/USE" :filter vhdl-insert-library-use-menu)
    "-----"
    ("Right Indent Region"
     ["1 Level"             (vhdl-indent-region (mark) (point) 1) (mark)]
     ["2 Levels"            (vhdl-indent-region (mark) (point) 2) (mark)]
     ["3 Levels"            (vhdl-indent-region (mark) (point) 3) (mark)]
     ["4 Levels"            (vhdl-indent-region (mark) (point) 4) (mark)])
    ("Left Indent Region"
     ["1 Level"             (vhdl-dedent-region (mark) (point) 1) (mark)]
     ["2 Levels"            (vhdl-dedent-region (mark) (point) 2) (mark)]
     ["3 Levels"            (vhdl-dedent-region (mark) (point) 3) (mark)]
     ["4 Levels"            (vhdl-dedent-region (mark) (point) 4) (mark)])
    ["Comment Region"       vhdl-comment-region    (mark)]
    ["Uncomment Region"     vhdl-uncomment-region  (mark)]
    "-----"
    ("Check VHDL Source"
     ["Behavioral Mode"     analyze-check-beh        t]
     ["RTL Mode"            analyze-check-rtl        t])
    ["Format VHDL Source"   vhdl-vfmt-filter         t]
    ["Format VHDL Region"   vhdl-vfmt-region    (mark)]
    ["Fix Keyword Case"     vhdl-fix-keyword-case    t]
    "-----"
    ["Print VHDL Buffer"    vhdl-print-buffer        t]
    ["Print VHDL Region"    vhdl-print-region   (mark)]
    "-----"
    ("Analyze" :filter vhdl-get-analyze-menu)
    ["Next Error"           analyze-next-error       t]
    ["Previous Error"       analyze-previous-error   t]
    "-----"
    ["Goto ..."             function-menu            t]
    "-----"
    ["About ..."            vhdl-show-version        t]))

(defvar vhdl-popup-menu
  '("VHDL"
    ("Include Header"
     ["VHDL File Header"    (vexpand (concat vhdl-header-dir "/FileHeader"))    t]
     "-----"
     ["Entity"              (vexpand (concat vhdl-header-dir "/EntityHeader"))  t]
     ["Architecture"        (vexpand (concat vhdl-header-dir "/ArchHeader"))    t]
     ["Configuration"       (vexpand (concat vhdl-header-dir "/ConfigHeader"))  t]
     ["Package"             (vexpand (concat vhdl-header-dir "/PackageHeader")) t]
     ["Process"             (vexpand (concat vhdl-header-dir "/ProcessHeader")) t]
     ["Subprogram"          (vexpand (concat vhdl-header-dir "/SubProgHeader")) t]
     "-----"
     ("Set Project Id" :filter vhdl-get-prj-id-menu))
    ("Insert Default"
     ["COMPONENT"           (vhdl-insert-default-component)         t]
     ["PORT MAP"            (vhdl-insert-default-port-map)          t]
     ["SIGNAL LIST"         (vhdl-insert-default-signals)           t]
     ["CONFIGURATION"       (vhdl-insert-default-configuration)     t])
    ("Insert LIBRARY/USE" :filter vhdl-insert-library-use-menu)
    "-----"
    ("Check VHDL Source"
     ["Behavioral Mode"     analyze-check-beh        t]
     ["RTL Mode"            analyze-check-rtl        t])
    ["Format VHDL Source"   vhdl-vfmt-filter         t]
    ["Fix Keyword Case"     vhdl-fix-keyword-case    t]
    ["Print VHDL Buffer"    vhdl-print-buffer        t]
    "-----"
    ("Analyze" :filter vhdl-get-analyze-menu)
    ["Next Error"           analyze-next-error       t]
    ["Previous Error"       analyze-previous-error   t]
    "-----"
    ["Goto ..."             function-menu            t]))


(defvar vhdl-region-menu
  '("VHDL Regions"
    ("Right Indent Region"
     ["1 Level"             (vhdl-indent-region (mark) (point) 1) t]
     ["2 Levels"            (vhdl-indent-region (mark) (point) 2) t]
     ["3 Levels"            (vhdl-indent-region (mark) (point) 3) t]
     ["4 Levels"            (vhdl-indent-region (mark) (point) 4) t])
    ("Left Indent Region"
     ["1 Level"             (vhdl-dedent-region (mark) (point) 1) t]
     ["2 Levels"            (vhdl-dedent-region (mark) (point) 2) t]
     ["3 Levels"            (vhdl-dedent-region (mark) (point) 3) t]
     ["4 Levels"            (vhdl-dedent-region (mark) (point) 4) t])
    "-----"
    ["Format VHDL Region"   vhdl-vfmt-region                      t]
    "-----"
    ["Comment Region"       vhdl-comment-region                   t]
    ["Uncomment Region"     vhdl-uncomment-region                 t]
    "-----"
    ["Print VHDL Region"    vhdl-print-region                     t]))



(defun vhdl-popup-menu (e)
  "Popup either vhdl-region-menu if a region is active and vhdl-popup-menu
   otherwise."
  (interactive "@e")
  (if (mark)
      (popup-menu vhdl-region-menu)
    (popup-menu vhdl-popup-menu)))



;;============================================================================;;
;;                          abbreviation table                                ;;
;;============================================================================;;

(defvar vhdl-mode-abbrev-table nil
  "Table of abbreviations use in vhdl-mode.")

;;if no abbreviation table exists or it is nil, create a new one.
(and (or (not (boundp 'vhdl-mode-abbrev-table))
         (not vhdl-mode-abbrev-table))
     (let ((keyword-abbrev-list   nil)
           (attribute-abbrev-list nil))

       ;; all keywords:
       (mapcar '(lambda (arg)
                  (let* ((hook (make-symbol (concat "vhdl-" arg "-hook")))
                         (hook (intern (symbol-name hook))))
                    (setq keyword-abbrev-list (cons (list arg "" hook 0) keyword-abbrev-list))))
               vhdl-keyword-list)

       ;; all attributes:
       (mapcar '(lambda (arg)
                  (let* ((hook (make-symbol (concat "vhdl-tick-" arg "-hook")))
                         (hook (intern (symbol-name hook))))
                    (setq attribute-abbrev-list (cons (list arg "" hook 0) attribute-abbrev-list))))
               vhdl-attribute-list)

       (define-abbrev-table 'vhdl-mode-abbrev-table
         (append keyword-abbrev-list
                 attribute-abbrev-list
                 '(("aa"              ""   vhdl-array-hook            0)
                   ("ac"              ""   vhdl-access-hook           0)
                   ("af"              ""   vhdl-after-hook            0)
                   ("ar"              ""   vhdl-architecture-hook     0)
                   ("al"              ""   vhdl-alias-hook            0)
                   ("as"              ""   vhdl-assert-hook           0)
                   ("at"              ""   vhdl-attribute-hook        0)
                   ("be"              ""   vhdl-begin-hook            0)
                   ("bl"              ""   vhdl-block-hook            0)
                   ("bo"              ""   vhdl-body-hook             0)
                   ("bu"              ""   vhdl-buffer-hook           0)
                   ("ca"              ""   vhdl-case-hook             0)
                   ("cm"              ""   vhdl-component-hook        0)
                   ("cn"              ""   vhdl-constant-hook         0)
                   ("co"              ""   vhdl-configuration-hook    0)
                   ("di"              ""   vhdl-disconnect-hook       0)
                   ("do"              ""   vhdl-downto-hook           0)
                   ("en"              ""   vhdl-entity-hook           0)
                   ("ef"              ""   vhdl-endif-hook            0)
                   ("eg"              ""   vhdl-endgenerate-hook      0)
                   ("el"              ""   vhdl-else-hook             0)
                   ("eo"              ""   vhdl-endloop-hook          0)
                   ("ep"              ""   vhdl-endprocess-hook       0)
                   ("es"              ""   vhdl-elsif-hook            0)
                   ("ex"              ""   vhdl-exit-hook             0)
                   ("fg"              ""   vhdl-for-generate-hook     0)
                   ("fi"              ""   vhdl-file-hook             0)
                   ("fu"              ""   vhdl-function-hook         0)
                   ("ge"              ""   vhdl-generic-hook          0)
                   ("gn"              ""   vhdl-generate-hook         0)
                   ("gr"              ""   vhdl-group-hook            0)
                   ("gu"              ""   vhdl-guarded-hook          0)
                   ("ig"              ""   vhdl-if-generate-hook      0)
                   ("it"              ""   vhdl-inertial-hook         0)
                   ("im"              ""   vhdl-impure-hook           0)
                   ("io"              ""   vhdl-inout-hook            0)
                   ("la"              ""   vhdl-label-hook            0)
                   ("li"              ""   vhdl-library-hook          0)
                   ("lk"              ""   vhdl-linkage-hook          0)
                   ("lo"              ""   vhdl-loop-hook             0)
				   ("lt"              ""   vhdl-literal-hook          0)
                   ("op"              ""   vhdl-open-hook             0)
                   ("ot"              ""   vhdl-others-hook           0)
                   ("pa"              ""   vhdl-package-hook          0)
                   ("pc"              ""   vhdl-procedure-hook        0)
                   ("po"              ""   vhdl-port-hook             0)
                   ("pp"              ""   vhdl-postponed-hook        0)
                   ("pr"              ""   vhdl-process-hook          0)
                   ("pu"              ""   vhdl-pure-hook             0)
                   ("ra"              ""   vhdl-range-hook            0)
                   ("rc"              ""   vhdl-record-hook           0)
                   ("re"              ""   vhdl-report-hook           0)
                   ("rg"              ""   vhdl-register-hook         0)
                   ("rj"              ""   vhdl-reject-hook           0)
                   ("rt"              ""   vhdl-return-hook           0)
                   ("se"              ""   vhdl-select-hook           0)
                   ("sh"              ""   vhdl-shared-hook           0)
                   ("si"              ""   vhdl-signal-hook           0)
                   ("sl"              "std_logic"           nil       0)
                   ("slv"             "std_logic_vector"    nil       0)
                   ("su"              ""   vhdl-subtype-hook          0)
                   ("suv"             "std_ulogic_vector"   nil       0)
                   ("sv"              ""   vhdl-severity-hook         0)
                   ("th"              ""   vhdl-then-hook             0)
                   ("tr"              ""   vhdl-transport-hook        0)
                   ("ty"              ""   vhdl-type-hook             0)
                   ("ua"              ""   vhdl-unaffected-hook       0)
                   ("un"              ""   vhdl-until-hook            0)
                   ("va"              ""   vhdl-variable-hook         0)
                   ("wa"              ""   vhdl-wait-hook             0)
                   ("wu"              ""   vhdl-wait-until-hook       0)
                   ("wh"              ""   vhdl-when-hook             0)
                   ("wi"              ""   vhdl-while-hook            0)
                   ("wt"              ""   vhdl-with-hook             0)
                   ("xn"              ""   vhdl-xnor-hook             0)

                   ("acc"             ""   vhdl-access-hook           0)
                   ("ali"             ""   vhdl-alias-hook            0)
                   ("arch"            ""   vhdl-architecture-hook     0)
                   ("attr"            ""   vhdl-attribute-hook        0)
                   ("comp"            ""   vhdl-component-hook        0)
                   ("conf"            ""   vhdl-configuration-hook    0)
                   ("cons"            ""   vhdl-constant-hook         0)
                   ("disc"            ""   vhdl-disconnect-hook       0)
                   ("down"            ""   vhdl-downto-hook           0)
                   ("elseif"          ""   vhdl-elsif-hook            0)
                   ("endif"           ""   vhdl-endif-hook            0)
                   ("enti"            ""   vhdl-entity-hook           0)
                   ("forgen"          ""   vhdl-for-generate-hook     0)
                   ("foruse"          ""   vhdl-for-use-hook          0)
                   ("forl"            ""   vhdl-for-loop-hook         0)
                   ("forloop"         ""   vhdl-for-loop-hook         0)
                   ("func"            ""   vhdl-function-hook         0)
                   ("ifgen"           ""   vhdl-if-generate-hook      0)
                   ("libr"            ""   vhdl-library-hook          0)
                   ("low"             ""   vhdl-tick-low-hook         0)
                   ("pack"            ""   vhdl-package-hook          0)
                   ("post"            ""   vhdl-postponed-hook        0)
                   ("prod"            ""   vhdl-procedure-hook        0)
                   ("proc"            ""   vhdl-process-hook          0)
                   ("seve"            ""   vhdl-severity-hook         0)
                   ("sig"             ""   vhdl-signal-hook           0)
                   ("subt"            ""   vhdl-subtype-hook          0)
                   ("trans"           ""   vhdl-transport-hook        0)
                   ("vari"            ""   vhdl-variable-hook         0)
                   ("int"             "integer"               nil     0)
                   ("uns"             "unsigned"              nil     0)
                   ("sgn"             "signed"                nil     0)
                   ("stdl"            "std_logic"             nil     0)
                   ("stdlv"           "std_logic_vector"      nil     0)
                   ("stdu"            "std_ulogic"            nil     0)
                   ("stduv"           "std_ulogic_vector"     nil     0)
                   ("bitv"            "bit_vector"            nil     0))))))



;;============================================================================;;
;;                               syntax table                                 ;;
;;============================================================================;;

(defvar vhdl-mode-syntax-table nil
  "Syntax table in use in vhdl-mode buffers.")

(setq vhdl-mode-syntax-table (make-syntax-table))
(modify-syntax-entry ?-  ". 12" vhdl-mode-syntax-table)
(modify-syntax-entry ?\n ">"    vhdl-mode-syntax-table)
(modify-syntax-entry ?_  "w"    vhdl-mode-syntax-table)
(modify-syntax-entry ?#  "_"    vhdl-mode-syntax-table)
(modify-syntax-entry ?$  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?+  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?*  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?/  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?&  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?|  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?<  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?=  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?>  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?.  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?:  "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?\\ "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?\; "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?\[ "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?\] "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?\{ "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?\} "."    vhdl-mode-syntax-table)
(modify-syntax-entry ?\" "\""   vhdl-mode-syntax-table)
(modify-syntax-entry ?\' "\'"   vhdl-mode-syntax-table)
(modify-syntax-entry ?\( "()"   vhdl-mode-syntax-table)
(modify-syntax-entry ?\) ")("   vhdl-mode-syntax-table)
(modify-syntax-entry ?\" "\""   vhdl-mode-syntax-table)



;;============================================================================;;
;;                               key mappings                                 ;;
;;============================================================================;;

(defvar vhdl-mode-map nil
  "Keymap used in VHDL mode.")

(let ((map (make-sparse-keymap)))
  (define-key map '(backspace)               'backward-delete-char)
  (define-key map '(delete)                  'delete-char)
  (define-key map '(tab)                     'vhdl-indent-line)
  (define-key map '(shift tab)               'tab-to-tab-stop)
  (define-key map '(return)                  'vhdl-newline)
  (define-key map " "                        'vhdl-space)
  (define-key map "'"                        'vhdl-single-quote)
  (define-key map "."                        'vhdl-period)
  (define-key map ","                        'vhdl-comma)
  (define-key map ";"                        'vhdl-semicolon)
  (define-key map "-"                        'vhdl-dash)
  (define-key map "+"                        'vhdl-punctuation)
  (define-key map "="                        'vhdl-punctuation)
  (define-key map ":"                        'vhdl-punctuation)
  (define-key map "("                        'vhdl-punctuation)
  (define-key map ")"                        'vhdl-paren-close)
  (define-key map "["                        'vhdl-punctuation)
  (define-key map "]"                        'vhdl-punctuation)
  (define-key map '(button3)                 'vhdl-popup-menu)
  (define-key map '(f6)                      'analyze-check-noquery)
  (define-key map '(shift f6)                'analyze-next-error)
  (define-key map '(control f6)              'analyze-previous-error)
  (define-key map '(f7)                      'analyze-noquery)
  (define-key map '(shift f7)                'analyze-next-error)
  (define-key map '(control f7)              'analyze-previous-error)
  (define-key map [(control c) (control a)]  'analyze)
  (define-key map [(control c) (control b)]  'analyze-check-beh)
  (define-key map [(control c) (control c)]  'vhdl-comment-region)
  (define-key map [(control c) (control f)]  'vhdl-vfmt-filter)
  (define-key map [(control c) (f)]          'vhdl-vfmt-region)
  (define-key map [(control c) (control k)]  'vhdl-fix-keyword-case)
  (define-key map [(control c) (control r)]  'analyze-check-rtl)
  (define-key map [(control c) (control p)]  'vhdl-print-buffer)
  (define-key map [(control c) (control u)]  'vhdl-uncomment-region)
  (define-key map [(control c) (control v)]  'vhdl-show-version)

  ;; HP specials
  (if (string-match "Hewlett-Packard" (x-server-vendor))
      (progn
        (define-key map '(hpBackTab)         'tab-to-tab-stop)
        (define-key map '(hpDeleteChar)      'delete-char)
        (define-key map '(hpInsertChar)      'overwrite-mode)))

  (setq vhdl-mode-map map))



(defvar vhdl-minibuffer-map nil
  "Keymap to use when reading from the minibuffer in VHDL-mode.")

(let ((map (make-sparse-keymap)))
  (define-key map '(backspace) 'backward-delete-char)
  (define-key map '(delete) 'delete-char)
  (define-key map "\C-g" 'abort-recursive-edit)
  (define-key map "\r"   'vhdl-newline)
  (define-key map "\n"   'vhdl-newline)
  (define-key map " "    'vhdl-space)
  (define-key map "'"    'vhdl-single-quote)
  (define-key map "."    'vhdl-period)
  (define-key map ","    'vhdl-comma)
  (define-key map ";"    'vhdl-semicolon)
  (define-key map "-"    'vhdl-punctuation)
  (define-key map "+"    'vhdl-punctuation)
  (define-key map "="    'vhdl-punctuation)
  (define-key map ":"    'vhdl-punctuation)
  (define-key map "("    'vhdl-punctuation)
  (define-key map ")"    'vhdl-punctuation)
  (define-key map "["    'vhdl-punctuation)
  (define-key map "]"    'vhdl-punctuation)
  (setq vhdl-minibuffer-map map))


;; we must use the same syntax-table for minibuffers as in vhdl-buffers
;; in order to prevent problems with expansion of identifiers like
;; test_conf to test_CONFIGURATION and the like !!

(add-hook 'minibuffer-setup-hook
          '(lambda nil (set-syntax-table vhdl-mode-syntax-table)))



(defvar vhdl-mode-hook nil
  "List of functions to be executed upon entering VHDL mode.")


;;============================================================================;;
;;                                  toolbar                                   ;;
;;============================================================================;;

(defvar vhdl-toolbar-icon-directory toolbar-icon-directory)

(if (and (featurep 'toolbar) (not vhdl-inhibit-toolbar))
    (progn
      (defvar vhdl::toolbar-open-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-open.xpm" "vhdl-open.xbm")
                           vhdl-toolbar-icon-directory))
        "vhdl-open")
      (defvar vhdl::toolbar-save-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-save.xpm" "vhdl-save.xbm")
                           vhdl-toolbar-icon-directory))
        "vhdl-save")
      (defvar vhdl::toolbar-cut-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-cut.xpm" "vhdl-cut.xbm")
                           vhdl-toolbar-icon-directory))
        "vhdl-cut")
      (defvar vhdl::toolbar-copy-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-copy.xpm" "vhdl-copy.xbm")
                           vhdl-toolbar-icon-directory))
        "vhdl-copy")
      (defvar vhdl::toolbar-paste-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-paste.xpm" "vhdl-paste.xbm")
                           vhdl-toolbar-icon-directory))
        "vhdl-paste")
      (defvar vhdl::toolbar-undo-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-undo.xpm" "vhdl-undo.xbm")
                           vhdl-toolbar-icon-directory))
        "vhdl-undo")
      (defvar vhdl::toolbar-print-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-printer.xpm" "vhdl-printer.xbm")
                           vhdl-toolbar-icon-directory))
        "vhdl-print")
      (defvar vhdl::toolbar-vfmt-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-vfmt.xpm" "vhdl-vfmt.xbm")
                           vhdl-toolbar-icon-directory))
        "vfmt")
      (defvar vhdl::toolbar-analyze-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-analyze.xpm" "vhdl-analyze.xbm")
                           vhdl-toolbar-icon-directory))
        "analyze")
      (defvar vhdl::toolbar-check-rtl-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-check-rtl.xpm" "vhdl-check-rtl.xbm")
                           vhdl-toolbar-icon-directory))
        "check-rtl")
      (defvar vhdl::toolbar-check-beh-icon
        (toolbar-make-button-list
         (expand-file-name (if (featurep 'xpm) "vhdl-check-beh.xpm" "vhdl-check-beh.xbm")
                           vhdl-toolbar-icon-directory))
        "check-beh")
      
      
      (defvar vhdl::toolbar
        '([vhdl::toolbar-open-icon
           find-file  t
           "Open a file"]
          [vhdl::toolbar-save-icon
           save-buffer  t
           "Save buffer"]
          [vhdl::toolbar-print-icon
           vhdl-print-buffer  t
           "Print VHDL buffer"]
          [vhdl::toolbar-cut-icon
           x-kill-primary-selection  t
           "Kill region"]
          [vhdl::toolbar-copy-icon
           x-copy-primary-selection  t
           "Copy region"]
          [vhdl::toolbar-paste-icon
           x-yank-clipboard-selection  t
           "Paste from clipboard"]
          [vhdl::toolbar-undo-icon
           undo  t
           "Undo edit"]
          [vhdl::toolbar-vfmt-icon
           vhdl-vfmt-filter t
           "Format buffer"]
          [vhdl::toolbar-analyze-icon
           analyze-buffer-with-dialog t 
           "Analyse buffer"]
          [vhdl::toolbar-check-beh-icon
           analyze-check-beh-with-dialog  t 
           "Coding Standard Check Behavior"]
          [vhdl::toolbar-check-rtl-icon
           analyze-check-rtl-with-dialog  t 
           "Coding Standard Check RTL"]
          ))
      )
  )

;;============================================================================;;
;;                               mode definition                              ;;
;;============================================================================;;

(defun vhdl-mode ()
  "This is a mode intended to support program development in VHDL.  Most
control constructs and declarations of vhdl can be inserted in the buffer
by typing the first word of the construct you wish to enter, followed
by a space.
Other support includes indentation control and commenting.

The following variables can be set to configure the vhdl-mode:

   `vhdl-indent' : defaults to 4 and defines the number of columns to indent
                 syntactic levels in VHDL-Mode

   `vhdl-start-comment-column' : defaults to 40 and defines the column
                 to which trailing comments will be aligned to.

   `vhdl-end-comment-column' : defaults to 80 and defines the column
                 at which long comment lines will be wrapped.

   `vhdl-comment-region-prefix' : defaults to \"--+\" and defines the
                 comment string used to indicate a temporary comment
                 for region comment and uncomment

   `vhdl-keyword-case-defun' : defaults to 'upcase and controls the
                 character case of inserted VHDL keywords.
                 possible values are: 'upcase 'downcase and 'capitalize

   `vhdl-identifier-case-defun' : defaults to 'downcase and controls
                 character case of inserted VHDL identifier.
                 possible values are: 'upcase 'downcase and 'capitalize

   `vhdl-attribute-case-defun' : defaults to 'upcase and controls the
                 character case of inserted VHDL attributes.
                 possible values are: 'upcase 'downcase and 'capitalize

   `vhdl-abbrev-mode' : defaults to t. Turn on/off abbreviations.
                 On if `vhdl-abbrev-mode' is *non* nil.

   `vhdl-configuration-suffix' : defaults to \"_cfg\" and defines the
                 suffix to use for the default configuration name

   `vhdl-formatting-command' : defaults to \"vfmt\" and defines which
                 program is used to format VHDL buffer or region

   `vhdl-format-region-arg' : defaults to \"-k\" and defines the argument
                 used when calling the `vhdl-formatting-command' on regions

   `vhdl-check-command' : defaults to \"vlint\" and defines which
                 program is used to check the code of the VHDL buffer

   `vhdl-print-command' : defaults to \"vpr\" and defines the command
                 used to print vhdl buffers and regions

   `vhdl-project-id-list' defaults to \"no menu\" and configures
                 the VHDL menu -> Include HEADER -> Set Project Id menu
                 This menu sets the environment variable VDS_PROJECTID
                 to the selected menu entry. Used by the Include HEADER
                 functions.
                 Example:
                 (setq vhdl-project-id-list '(\"asic1\" \"asic2\"))

   `vhdl-library-use-list' : This list defines the submenu 'LIBRARY/USE'
                 in the 'VHDL' menu for detailed info and default value
                 use the describe-variable facility
                 Syntax:  '( ( <menustring> <library> <use> ... ) ... )

   `vhdl-inhibit-toolbar' : defaults to nil and controls the appearance
                 of the specialized Info toolbar. A value other than nil
                 will inhibit the toolbar

   `vhdl-toolbar-icon-directory' : defaults to toolbar-icon-directory and
                 determines the location of the vhdl-toolbar icons

   `vhdl-file-search-path' : defaults to '(\".\" \"./sources\")
                 This searchpath is used by the \"Insert Default\"
                 functions and determines which directories are scanned
                 to find the file of an entity.
                 Example:
                 (setq vhdl-file-search-path '(\".\" \"./vhdl\"))

   `vhdl-file-suffixes' : defaults to \".vhdl:.vhd\" this colon separated
                 list is used by the \"Insert Default\" functions to find
                 file <entity><suffix>

   `analyze-command' : defaults to \"cv -m\" and defines the default
                 setting used by pushing the analyze button in the toolbar

   `analyze-command-list' : defaults to nil and defines the list of menu
                 entries to start the VHDL analyzer.
                 Example:
                 (setq analyze-command-list '(\"cv -m\" \"cv -m -comp\"))

   `vhdl-mode-map' : is the keymap of the vhdl-mode

   `vhdl-mode-hook' : This hook is called when entering the vhdl-mode
                 Example:

                 (defun my-vhdl-startup ()
                    (setq vhdl-indent 3)
                    ...
                    ...)
                 (add-hook 'vhdl-mode-hook 'my-vhdl-startup)
"
  (interactive)

  (kill-all-local-variables)

  (setq major-mode 'vhdl-mode)
  (setq mode-name "VHDL")

  ;; features required by vhdl-mode
  (require 'analyze)
  (require 'vexpand)

  ;; mode setup
  (use-local-map vhdl-mode-map)

  (setq local-abbrev-table vhdl-mode-abbrev-table)
  (abbrev-mode 1)
  (set-syntax-table vhdl-mode-syntax-table)

  ;; Always use space characters for indentation rather than tabs
  (make-local-variable 'indent-tabs-mode)
  (setq indent-tabs-mode nil)

  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'vhdl-indent-line)
  
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)

  (make-local-variable 'fume-menubar-menu-name)
  (setq fume-menubar-menu-name "Goto...")

  ;; set location to nil means default position at end of menu
  (make-local-variable 'fume-menubar-menu-location)
  (setq fume-menubar-menu-location nil)

  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments t)
  ;; Syntax of VHDL comments (used by comment-region)
  (make-local-variable 'comment-start)
  (setq comment-start "-- ")
  (make-local-variable 'comment-end)
  (setq comment-end "")

  ;; Define the starting string for comments (like comment-start, but also matches
  ;; spaces so that following comment lines will have the same layout).
  ;; This is used by auto-fill mode.
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "--+ *")

  ;; Define the starting column for right-of-code comments
  ;; (Used by auto-fill mode)
  (setq comment-column (- vhdl-start-comment-column 1))
  (make-local-variable 'vhdl-end-comment-column)

  ;; Install the font-lock keywords list
  ;; and don't care about the keyword case
  (put 'vhdl-mode 'font-lock-keywords 'vhdl-fontlock-keywords)
  (put 'vhdl-mode 'font-lock-keywords-case-fold-search t)

  ;; remember entity and architecture names to offer them as defaults
  ;; in the appropriate situations.
  (make-local-variable 'vhdl-saved-entity-name)
  (make-local-variable 'vhdl-saved-architecture-name)
  
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'vhdl-comment-indent)

  (and current-menubar
       (not (assoc (car vhdl-pulldown-menu) current-menubar))
       (progn
         (set-buffer-menubar (copy-sequence current-menubar))
         (add-submenu nil vhdl-pulldown-menu)))

  (run-hooks 'vhdl-mode-hook)

  ;; this must be done after running vhdl-mode-hook because
  ;; vhdl-identifier-case-defun might have been modified:
  (vhdl-get-identifier-buffer-defun)

  ;; enable toolbar
  (if (and (featurep 'toolbar) (not vhdl-inhibit-toolbar))
      (set-specifier default-toolbar (cons (current-buffer) vhdl::toolbar)))

  (abbrev-mode 0))



(defun vhdl-get-identifier-buffer-defun ()
  "Set the value of vhdl-identifier-case-buffer-defun to something
   appropriate for modifying buffer contents."
  (setq vhdl-identifier-case-buffer-defun
        (cond ((eq vhdl-identifier-case-defun 'upcase)
               'upcase-word)
              ((eq vhdl-identifier-case-defun 'downcase)
               'downcase-word)
              ((eq vhdl-identifier-case-defun 'capitalize)
               'capitalize-word)
              (t
               'downcase-word))))

;;===========================================================================;;
;;                        freeze display function                            ;;
;;===========================================================================;;

(defvar vhdl-current-line 0)
(defvar vhdl-current-center-window-line 0)

(defun vhdl-save-window-pos ()
  (let ((curpos (point)))
    (beginning-of-line)
    (setq vhdl-current-line (count-lines (point-min) (point)))
    (move-to-window-line nil)
    (setq vhdl-current-center-window-line (count-lines (point-min) (point)))
    (goto-char curpos)))

(defun vhdl-restore-window-pos ()
  (goto-line (1+ vhdl-current-center-window-line))
  (recenter)
  (goto-line (1+ vhdl-current-line))
)

;;===========================================================================;;
;;                            external filter tools                          ;;
;;===========================================================================;;

(defun vhdl-check-stderr-file (filename progname)
  "Check for errors or warnings. If filesize greater than zero report at modeline.
   If the filesize is zero delete the file silently"
  (let ((attr (file-attributes filename)))
    (if attr
        (if (not (zerop (nth 7 attr)))
            (error (concat progname " writes something to stderr. Check file " filename))
          (delete-file filename)))))

(defun vhdl-vfmt-region (start end)
  "Run the current buffer through the vfmt formatter."
  (interactive "r")
  (message "formatting ...")
  (vhdl-save-window-pos)
  (call-process-region start end vhdl-formatting-command t '(t "vfmt-stderr.log") nil vhdl-format-region-arg)
  (vhdl-restore-window-pos)
  (message "formatting ... done")
  (vhdl-check-stderr-file "vfmt-stderr.log" vhdl-formatting-command))

(defun vhdl-vfmt-filter ()
  "Run the current buffer through the vfmt formatter."
  (interactive)
  (if (y-or-n-p (concat "Format buffer '" (buffer-name) "' ? "))
      (progn
        (message "formatting ...")
        (vhdl-save-window-pos)
        (call-process-region (point-min) (point-max) vhdl-formatting-command t '(t "vfmt-stderr.log") nil)
        (vhdl-restore-window-pos)
        (message "formatting ... done")
        (vhdl-check-stderr-file "vfmt-stderr.log" vhdl-formatting-command))
    (error "formatting cancelled ...")))



(defun vhdl-print-buffer ()
  "Print the current buffer using the vpr print tool."
  (interactive)
  (if (y-or-n-p (concat "Print buffer '" (buffer-name) "' ? "))
      (progn
        (message "spooling ...")
        (call-process-region (point-min) (point-max) 
                             "vpr" nil t nil
                             "-N" (or buffer-file-name (buffer-name)))
        (message "spooling ... done"))
    (error "printing cancelled ...")))



(defun vhdl-print-region (start end)
  "Print the current buffer using the vpr print tool."
  (interactive "r")
  (if (y-or-n-p (concat "Print selected region of '" (buffer-name) "' ? "))
    (progn
      (message "spooling ...")
      (call-process-region start end 
                           "vpr" nil t nil "-N" 
                           (concat "'" (buffer-name) " (partial buffer)'"))
      (message "spooling ... done"))
    (error "printing cancelled ...")))



(defun vhdl-fix-keyword-case ()
  "Ensure key words and names of predefined attributes are in upper case,
   and all user-defined names are in lower case."
  (interactive)
  (message "formatting ...")
  (save-excursion
    (vhdl-save-window-pos)
    (call-process-region (point-min) (point-max) "vcase" t t nil)
    (vhdl-restore-window-pos))
  (message "formatting ... done"))




;;===========================================================================;;
;;                             predicate functions                           ;;
;;===========================================================================;;


(defun vhdl-in-quote-p ()
  "Check if point is inside a quoted string."
  (let ((opoint (point)))
    (save-excursion ; preceeded by odd number of quotes?
      (beginning-of-line)
      (re-search-forward "^\\([^\"]*\"[^\"]*\"\\)*" opoint t)
      (search-forward "\"" opoint t))))



(defun vhdl-in-comment-p ()
  "Check if point is to right of comment delimiter."
  (let ((opoint (point)))
    (save-excursion ; finds an unquoted comment
      (beginning-of-line)
      (re-search-forward "^\\([^\"]*\"[^\"]*\"\\)*[^\"]*--" opoint t))))


(defun vhdl-first-on-line-p ()
  "Returns t if only whitespace characters are between bol and point."
  (save-excursion
    (let ((opoint (point)))
      (beginning-of-line)
      (not (re-search-forward "[^\\s-]" opoint t)))))



(defun vhdl-looking-at-eol-p ()
  "Returns t if only optional whitespace characters are on line."
  (save-excursion
    (beginning-of-line)
    (looking-at "\\s-*$")))



(defun vhdl-backward-looking-at (regexp)
  (let ((begin (point))
        (found (re-search-backward regexp nil t)))
    (goto-char begin)
    (and found (= begin (match-end 0)))))


;;===========================================================================;;
;;                             prompt functions                              ;;
;;===========================================================================;;


(add-hook 'minibuffer-setup-hook
          '(lambda nil (setq local-abbrev-table vhdl-mode-abbrev-table)))

(defun vhdl-read-string (prompt &optional initial-contents)
  "Cloned version of read-string from minibuf.el with special keymap."
  (let ((minibuffer-completion-table nil))
      (read-from-minibuffer prompt initial-contents vhdl-minibuffer-map)))



(defun vhdl-prompt (prompt &optional following-string)
  "PROMPT for string and insert it in buffer with optional FOLLOWING-STRING.
   Prompt may contain default value indicated as prompt=default which will be used
   when no value is provided."
  (let ((old-point (point))
        (def-value))
    (if (string-match "^\\(.*\\)=\\(.*\\)\\s-*$" prompt)
        (progn
          (setq def-value (substring prompt (match-beginning 2) (match-end 2)))
          (setq prompt    (substring prompt (match-beginning 1) (match-end 1)))))
    (insert "{" prompt "}")
    (let ((string (vhdl-read-string (concat prompt " ? ") def-value)))
      (discard-input)
      (if (and (string-match "^\\s-*$" string) def-value)
          (setq string def-value))
      (delete-region old-point (point))
      (insert string)
      (vhdl-insert-keyword (or following-string ""))
      string)))



(defun vhdl-read (prompt)
  "Function similar to vhdl-prompt, but unlike vhdl-prompt vhdl-read does not
   insert the prompted value but merely returns it for postprocessing."
  (let ((old-point (point)))
    (insert "{" prompt "}")
    (let ((string (downcase (vhdl-read-string (concat prompt " ? ")))))
      (delete-region old-point (point))
      string)))



(defun vhdl-get-parameter-list ()
  "Read from user a procedure or function argument list.
   Add parens unless arguments absent, and insert into buffer.
   Individual arguments are arranged vertically if entered one-at-a-time.
   Arguments ending with semicolon or comma are presumed single and stacked."
  (insert " (")
  (let (args (margin (current-column)))
    (if (string-equal (setq args (vhdl-prompt "parameter1")) "")
        (backward-delete-char 2)
      (progn
        (while (string-match "[,;]$" args)
          (newline)
          (indent-to margin)
          (setq args (vhdl-prompt "next_parameter")))
        (insert ")")))))



;;===========================================================================;;
;;                          indentation functions                            ;;
;;===========================================================================;;

(defconst vhdl-interface-list-regexp
  "\\(map\\|port\\|generic\\)\\W"
  "regular expression to decide whether the list is interface or parameter")

(defconst vhdl-label-indent-regexp
  (concat "\\(block"
          "\\|for"
          "\\|loop"
          "\\|process"
          "\\|while"
          "\\)\\W")
  "Regular expression after a label which cause a right indentation.")

(defconst vhdl-right-indent-regexp
  (concat "\\(architecture"
          "\\|begin"
          "\\|case"
          "\\|component"
          "\\|configuration"
          "\\|else"
          "\\|elsif"
          "\\|entity"
          "\\|for"
          "\\|function"
          "\\|generic"
          "\\|if"
          "\\|package"
          "\\|procedure"
          "\\|process"
          "\\|record"
          "\\|units"
          "\\|then"
          "\\|type"
          "\\|when"
          "\\)\\W")
  "Regular expression which cause a right indentation.")

(defconst vhdl-terminate-regexp
  (concat "\\(begin"
          "\\|is"
          "\\|loop"
          "\\|then"
          "\\)\\W")
  "if Regular expression is found at the end of a line stop scanning backward")

(defconst vhdl-left-indent-regexp
  (concat "\\(begin"
          "\\|else"
          "\\|elsif"
          "\\|end"
          "\\|then"
          "\\|when"
          "\\)\\W")
  "Regular expression which cause a left indentation.")

(defun vhdl-backward-to-indent ()
  (backward-to-indentation 1)
  (while (and (looking-at "[\n-]") (not (bobp)))
    (backward-to-indentation 1)))

(defun vhdl-last-char-in-prev-line (regexp)
  (save-excursion
    (forward-comment (- (buffer-size)))
    (if (bobp)
        t
      (backward-char 1)
      (looking-at regexp))))

(defun vhdl-next-keyword-is (regexp)
  (save-excursion
    (forward-sexp)
    (forward-sexp)
    (backward-sexp)
    (looking-at regexp)))

(defun vhdl-last-column-has-trailing-comment ()
  (save-excursion
      (if (forward-comment -1)
          (if (looking-at "--")
              (if (>= (current-column) (- comment-column vhdl-indent))
                  (current-column))))))

(defun vhdl-guess-last-statement ()
  "scan backward and find the actual indent"
  (save-excursion
    (let ((loop t)
          (first (point-max))
          save
          (last nil)
          (case-fold-search t))
      (while loop
        (backward-sexp)
        (if (bobp)
            (setq save (cons 'vhdl-new-statement 0)
                  loop nil))
        (if (< (point) first)
            (save-excursion
              (back-to-indentation)
              (setq first (point))))
        (if (= (point) first)
            (cond
             ((looking-at "case")
              (setq save (cons 'vhdl-case-statement
                               (current-column))
                    loop nil))
             ((looking-at vhdl-right-indent-regexp)
              (if last
                  (setq save last)
                (setq save (cons 'vhdl-indent-statement
                                 (current-column))))
              (setq loop nil))
             ((looking-at vhdl-left-indent-regexp)
              (if (and (looking-at "end")
                       (vhdl-next-keyword-is "record"))
                  (setq save (cons 'vhdl-new-statement
                                   (- (current-column) vhdl-indent)))
                (setq save (cons 'vhdl-new-statement
                                 (current-column))))
              (setq loop nil))
             ((and (not (looking-at vhdl-keyword-regexp))
                   (vhdl-next-keyword-is vhdl-label-indent-regexp))
              (if last
                  (setq save last)
                (setq save (cons 'vhdl-indent-statement
                                 (current-column))))
              (setq loop nil))
             ((save-excursion
                (backward-sexp)
                (looking-at vhdl-terminate-regexp))
              (setq save (cons 'vhdl-new-statement (current-column))
                    loop nil))
             ((or (vhdl-last-char-in-prev-line ";")
                  (bobp))
              (setq save (cons 'vhdl-new-statement (current-column))
                    loop nil))
             (t (setq last (cons 'vhdl-new-statement (current-column)))))))
      save)))


(defun vhdl-guess-syntax ()
  "Find out in which construct we are in the actual buffer; returns symbol/indent pair"
  (interactive)
  (setq parse-sexp-ignore-comments t)
  (let ((case-fold-search t)
        save-pos
        paren-pos
        state
        last-guess
        terminated
        paren-col
        prev-col)
    (save-excursion
      (setq save-pos (point))
      (setq state (parse-partial-sexp (point-min) (point)))
      ;; state is set as follows:
	  ;; The result is a list of eight elements describing the final state
	  ;; of the parse:
      ;;0. The depth in parentheses, counting from 0.
	  ;;1. The character position of the start of the innermost
	  ;;   parenthetical grouping containing the stopping point; `nil'
	  ;;   if none.
	  ;;2. The character position of the start of the last complete
	  ;;   subexpression terminated; `nil' if none.
	  ;;3. Non-`nil' if inside a string.  More precisely, this is the
	  ;;   character that will terminate the string.
	  ;;4. `t' if inside a comment (of either style).
	  ;;5. `t' if point is just after a quote character.
	  ;;6. The minimum parenthesis depth encountered during this scan.
	  ;;7. `t' if inside a comment of style "b".
	 
      (cond
       ;;
       ;; check for string
       ;;
       ((nth 3 state) ;; the string flag is set
        (cons 'vhdl-string 0)) ;; so, the return value is this list
       ;;
       ;; check for comment
       ;;
       ((and (looking-at "--")
             (>= (current-column) (- comment-column vhdl-indent))
             (setq prev-col (vhdl-last-column-has-trailing-comment)))
        (cons 'vhdl-trailing-comment prev-col))
       ;;
       ;; check for parenlevel
       ;;
       ((<= 1 (nth 0 state)) ;; parentheses level >= 1 
        (save-excursion
          (backward-up-list (nth 0 state))
          (setq paren-pos (point))
          (setq paren-col (current-column))
          (backward-sexp)
          ;;
          ;; check for interface list
          ;;
          (if (looking-at vhdl-interface-list-regexp)
              (progn
                ;;
                ;; check for interface list end
                ;;
                (goto-char save-pos)
                (if (and (looking-at ")")
                         (= 1 (nth 0 state)))
                    (progn
                      (goto-char paren-pos)
                      (back-to-indentation)
                      (cons 'vhdl-interface-list-end (current-column)))
                  ;;
                  ;; check for interface list intro
                  ;;
                  (goto-char save-pos)
                  (vhdl-backward-to-indent)
                  (setq prev-col (current-column))
                  (if (<= (point) paren-pos)
                      (cons 'vhdl-interface-list-intro (+ prev-col vhdl-indent))
                    (cons 'vhdl-interface-list prev-col))))
            ;;
            ;; We are in a parameter list
            ;;   and check for parameter list end
            (goto-char save-pos)
            (if (and (looking-at ")")
                     (= 1 (nth 0 state)))
                (progn
                  (goto-char paren-pos)
                  (cons 'vhdl-parameter-list-end (current-column)))
              ;;
              ;; check for parameter list intro
              ;;
              (goto-char save-pos)
              (vhdl-backward-to-indent)
              (setq prev-col (current-column))
              (if (<= (point) paren-pos)
                  (cons 'vhdl-parameter-list-intro (1+ paren-col))
                (cons 'vhdl-parameter-list prev-col))))))
       ;;
       ;; not in parenthesis: this case handles the rest
       ;;
       (t (setq last-guess (vhdl-guess-last-statement))
          (setq terminated (vhdl-last-char-in-prev-line ";"))
          (cond
           ((eq (car last-guess) 'vhdl-case-statement)
            (if (looking-at "end")
                last-guess
              (cons 'vhdl-case-statement (+ (cdr last-guess) vhdl-indent))))
           ;;
           ;;  last statement is indent but terminated
           ;;
           ((and (eq (car last-guess) 'vhdl-indent-statement)
                 terminated)
            (if (looking-at vhdl-left-indent-regexp)
                (if (looking-at "when")
                    (cons 'vhdl-right-left-indent-statement
                          (cdr last-guess))
                  (cons 'vhdl-left-indent-statement
                        (- (cdr last-guess) vhdl-indent)))
              (cons 'vhdl-new-statement (cdr last-guess))))
           ;;
           ;;  last statement is indent and unterminated
           ;;
           ((eq (car last-guess) 'vhdl-indent-statement)
            (if (looking-at vhdl-left-indent-regexp)
                (cons 'vhdl-right-left-indent-statement
                      (cdr last-guess))
              (cons 'vhdl-right-indent-statement
                    (+ (cdr last-guess) vhdl-indent))))
           
           ((eq (car last-guess) 'vhdl-new-statement)
            (if (looking-at vhdl-left-indent-regexp)
                (cond
                 ((save-excursion
                    (if (looking-at "end")
                        (progn
                          (forward-sexp)
                          (forward-sexp)
                          (backward-sexp)
                          (looking-at "case"))))
                  (cons 'vhdl-double-left-indent-statement
                        (- (cdr last-guess) vhdl-indent vhdl-indent)))
                 (t
                  (cons 'vhdl-left-indent-statement
                        (- (cdr last-guess) vhdl-indent))))
              (if terminated
                  last-guess
                (cons 'vhdl-open-statement
                      (+ (cdr last-guess) vhdl-indent)))))))))))

(defun vhdl-indent-line ()
  "indent the actual line"
  (interactive)
  (let ((save-pos (- (point-max) (point)))
        guess)
    (back-to-indentation)
    (setq guess (vhdl-guess-syntax))
    (if (interactive-p)
        (message (symbol-name (car guess))))
    (if (> (current-column) (cdr guess))
        (delete-horizontal-space))
    (indent-to (cdr guess))
    (if (> (- (point-max) save-pos) (point))
        (goto-char (- (point-max) save-pos)))))


(defun vhdl-indent-region (start end &optional times)
  "indent region (vhdl-indent) spaces.  Accepts number of times to indent"
  (interactive "r\np")
  (indent-code-rigidly start end (* times vhdl-indent)))



(defun vhdl-dedent-region (start end &optional times)
  "dedent region (vhdl-indent) spaces.  Accepts number of times to dedent"
  (interactive "r\np")
  (indent-code-rigidly start end (* -1 times vhdl-indent)))

(defun vhdl-wrap-comment-line () ;18/08/99 by C.Mayer
  "Handles a newline within a comment line (block/non-block): wraps around
   all words/special signs that are beyond column margin and potentially adjusts
   the previous line's block comment delimiters, This also involves comment line
   filling if the line consists of homogeneous characters."
   (interactive)
   (let ((backup-pos (point)) ;where the line break was intended to happen
		comm-start-pos
		comm-start-col
		comm-stop-pos
		comm-stop-col
		lead-space
		has-trailing-comment
		need-trailing-delim
		homogeneous-char
		(less-space 0)
		(column-delta 0)
		(delta 0))
		
		; at first, we analyze the line in which the break should happen:
		; search from BOL to first '--' outside of quoted string. This is the beginning
		; (left edge) of our comment and it is required to compute the indendation pos
		; of the next comment line.
		; Note that .*? represents the _shortest_ possible match (non-greedy) !
		(beginning-of-line)
		(re-search-forward "\\(\".*?\".*?\\)*\\(--\\)" backup-pos t)
		(setq comm-start-pos (match-beginning 2)) ;get the 2nd matching subexpression (the --)
		
		;look out for a trailing comment delimiter. If we find one, we'll terminate 
		;the line to be a framed block comment after inserting the line break.
		(goto-char (+ comm-start-pos 2)) ;skip the leading -- ...
		(if (looking-at ".+\\(--\\)\\s-*$") ; Note: \\s- is WHITESPACE, $ is EOL.
											; we only recognize a trailing comment if there is
											; at least one character between the comment delimiters
			(progn ;we do have a trailing comment delimiter
				   (setq comm-stop-pos (match-beginning 1))
				   (setq has-trailing-comment t)
				   (setq less-space 2))
				   
			(progn ;we have no trailing comment
				   (end-of-line)
				   (setq comm-stop-pos (point)))
		)
		
		
		; here, we've isolated comm-start-pos and comm-stop-pos; examples (| is bol):
		; | blah "--"  -- comment 
		;              ^        ^
		; |--
		; ^  (both markers are identical)
		;
		; |---    |----   |-- --
		;  ^ ^     ^  ^    ^  ^ (in the last case, less-space is set to 2 rather than 0)
		
		
		; now, let's look inside the comment: how many leading spaces ?
		; is it a homogeneous comment line (all chars within it are identical) ?
	    (goto-char comm-start-pos)
	    (setq comm-start-col (current-column))
	    (forward-char 2) ;skip the leading --
	    ;(message "crange: %d -> %d  %d" comm-start-pos comm-stop-pos less-space)
	    (if (< (point) comm-stop-pos)
		  (progn
			(if (>= (- comm-stop-pos comm-start-pos) (- 5 less-space))
				(setq homogeneous-char (following-char)); homogeneous lines must be at least --x-- or --xxx
				(setq homogeneous-char nil))            ; here, line is too short (e.g. ---- or --==)

		    (while (< (point) comm-stop-pos) 
				(if homogeneous-char ;look whether the homogeneous still assumption holds true
					(if (not (equal (following-char) homogeneous-char))
						(setq homogeneous-char nil)));nope - detected a different char -> line is not homogeneous
			
				(if (not lead-space)
					(if (not (equal (following-char) ? ))
						(setq lead-space (- (- (point) comm-start-pos) 2))))
					
				(forward-char 1)
			) ; for all characters in the comment line
		  )
		)
		(setq comm-stop-col (current-column))
		
		(if (equal homogeneous-char ?-)
		  (progn ; oops- this is a line consiting of purely - signs, so we have to re-think our
				 ; 'has-trailing-comment' decision
			(if (not vhdl-frame-block-comments)
			  (progn ;in fact, the user does not want framed block comments, so we
					 ; interpret this '-' - line as a non-framed one.
				(setq has-trailing-comment nil)
				(setq less-space 2)
			  )
			)
		  )
		)

		;(if homogeneous-char
		;(message "HOMOGENEOUS crange: %d (%d) -> %d (%d), %d" comm-start-pos comm-start-col comm-stop-pos comm-stop-col less-space)
		;(message "crange: %d (%d) -> %d (%d), %d" comm-start-pos comm-start-col comm-stop-pos comm-stop-col less-space))
	    ; now that we know everything about our comment line, let's find a valid 
	    ; point where to insert the newline...
	    (if homogeneous-char
			(progn ;eh - insert \n within or past homogeneous comment line
				   ;here, we exactly fill & align the line to the left
				   
				 ;delta > 0 means: line too short by this many chars
				 (setq delta (- (- vhdl-end-comment-column less-space) comm-stop-col))
				 (if (< (+ comm-stop-col delta) comm-start-col) (setq delta 0)) ;oops - no space left
				 (setq comm-stop-pos (+ comm-stop-pos delta))
				 (setq comm-stop-col (+ comm-stop-col delta))

				 ;we're still at comm-stop-col, just before the tailing -- (if any)
				 (if (> delta 0) (insert-char homogeneous-char delta))
				 (if (< delta 0) (delete-char delta)) ;deletes to the left ! (neg. delta)
				 (goto-char (+ comm-stop-pos less-space)) ;insert line break hereafter
				 (setq lead-space 2) ; the default: add two leading spaces after the new comment delim
				 (delete-horizontal-space) ;get rid of any trailing whitespace...
				 (setq need-trailing-delim t)
			)
			
			;return within 'real' (non-homogeneous) comment line
			(if (> comm-start-col vhdl-end-comment-column)
			  (progn ; the _start_ column of the entire comment is too far to the right
				(setq comm-start-col vhdl-indent) ; wrap next comment line at one indent
				(setq column-delta (- backup-pos comm-start-pos))
				(goto-char comm-start-pos)
				(delete-char 2) ;eat up the comment intro (will be re-inserted below)
				(delete-horizontal-space) ;get rid of any surrounding whitespace...
			  )
			  
			  ;here, the comm-start-pos is small enough. This should be the normal case.
			  (goto-char backup-pos); go back to where the user pressed return
			  ; step back (left) until we found legal point-of-linebreak

			;(message "drange: %d (%d) -> %d (%d), pos %d (%d), ls %s, cd %d delta %d" comm-start-pos comm-start-col
			; comm-stop-pos comm-stop-col (point) (current-column) lead-space column-delta delta)
			 (setq column-delta (current-column))
			 (while (> (current-column) (- vhdl-end-comment-column less-space)) (forward-word -1))
			; (message "erange: %d (%d) -> %d (%d), pos %d (%d), ls %s, cd %d delta %d" comm-start-pos comm-start-col
			; comm-stop-pos comm-stop-col (point) (current-column) lead-space column-delta delta)
			 ;OK - here, point sits on the location where the line break should happen
			 (setq column-delta (- column-delta (current-column))) ;remember how far we went back
			 (delete-horizontal-space) ;get rid of any surrounding whitespace...
			 (if has-trailing-comment ;pad with spaces & add trailing comment in upper line
			   (progn
				 (insert-char ? (- vhdl-end-comment-column (+ (current-column) 2))) ; add padding spaces
				 (insert "--") ; add comment trailer
				 (if (not (looking-at ".*--[ ]*$")) (setq need-trailing-delim t))
			   )
			 )
		   )
         )

		;here, point sits at the insertion place for our newline
		;(message "crange: %d (%d) -> %d (%d), pos %d (%d), ls %s, cd %d delta %d" comm-start-pos comm-start-col
		; comm-stop-pos comm-stop-col (point) (current-column) lead-space column-delta delta)
		(insert "\n")
		(insert-char ? comm-start-col) ; add leading spaces
		(insert "--") ; add comment leader
		(insert-char ? lead-space) ;and align leading space within comment block
		(forward-char column-delta) ;jump back onto where we previously were
		(if has-trailing-comment ;append the trailing comment delimiters in the new line as well
		  (progn 
			(setq backup-pos (point)) ;remember our cursor position
			(end-of-line) ;;and fix the eol: (potentially) add and align the trailing '--'
			(delete-horizontal-space)
			(if need-trailing-delim (insert "--")) ;there were no trailing -- in this line, yet.
			(forward-char -2) ;jump just-before the trailing --
			(delete-horizontal-space)
			(if (> (- vhdl-end-comment-column (+ (current-column) 2)) 0)
			  (insert-char ? (- vhdl-end-comment-column (+ (current-column) 2)))) ; add padding spaces
			
			(goto-char backup-pos) ;restore position
		  )
		)
))

;;===========================================================================;;
;;                        keys with special properties                       ;;
;;===========================================================================;;


(defun vhdl-expand-abbrev ()
  "Expand abbrev before point, if preceeding char is letter."
  (interactive)
  (or (bobp)                        ; will not expand at bob ...
      (vhdl-in-comment-p)             ; ... or within comments
      (vhdl-in-quote-p)               ; ... or within quoted strings
      (save-excursion               ; preceding char must be letter
        (backward-char 1)
        (not (looking-at "[A-Za-z]")))
      (expand-abbrev)))

(defun vhdl-space (count)
  "Self-insert space(s), indent a new (block) comment line if in (block) comment and past
    vhdl-end-comment-column."
  (interactive "p")
  (vhdl-expand-abbrev)
  (cond ((< (current-column) vhdl-end-comment-column)
         (self-insert-command count))
         
         ; here, we're past EOL
        ((vhdl-in-comment-p) ;we're inside a VHDL comment section
		  (vhdl-wrap-comment-line)
		  (insert-char ? count))
        (t ;we're not iside a comment but line wrap occurs, anyway.
         (self-insert-command count))))


(defun vhdl-dash (count)
  "--<any> starts a inline comment, --- starts a block comment"
  (interactive "p")
  (if (or (vhdl-in-comment-p)
          (vhdl-in-quote-p))
      (self-insert-command count)
    ;; turn overwriting off to prevent problems when deleting dash
    (overwrite-mode 0)
    (cond ((and abbrev-start-location (= abbrev-start-location (point)))
           (setq abbrev-start-location nil)
           (goto-char last-abbrev-location)
           (beginning-of-line nil)
           (vhdl-block-comment))
          ((/= (preceding-char) ?-) ; standard dash (minus)
           (self-insert-command count))
          (t
           (delete-char -1) ; delete the previous dash
           (let ((next-input (read-char)))
             (if (= next-input ?-) ; triple dash
                 (vhdl-block-comment)

               ;;;pushback the char
               (setq unread-command-event (character-to-event next-input (allocate-event)))
               (vhdl-inline-comment)))))))



(defun vhdl-newline ()
  "newline-and-indent or indent-new-comment-line if in comment
   and preceding character is a space."
  (interactive)
  (vhdl-expand-abbrev)
  (cond ((window-minibuffer-p (selected-window))
         (exit-minibuffer))
         
		((vhdl-in-comment-p) ;we're inside a VHDL comment section
			(vhdl-wrap-comment-line))     
         
        (t
         (newline-and-indent))))



(defun vhdl-punctuation (arg)
  "Perform abbreviation expansion and insert the punctuation character typed."
  (interactive "P")
  (vhdl-expand-abbrev)
  (self-insert-command (prefix-numeric-value arg)))



(defun vhdl-comma (arg)
  " ',,' --> '<='"
   (interactive "P")
   (vhdl-expand-abbrev)
   (if (or (/= (preceding-char) last-input-char)
           (vhdl-in-comment-p)
           (vhdl-in-quote-p))
       (self-insert-command (prefix-numeric-value arg))
     (progn
       (delete-char -1)
       (if (= (preceding-char) ? )
           (insert "<= ")
         (insert " <= ")))))



(defun vhdl-period (arg)
  " '..' --> '=>'"
  (interactive "P")
  (vhdl-expand-abbrev)
  (if (or (/= (preceding-char) last-input-char)
          (vhdl-in-comment-p)
          (vhdl-in-quote-p))
      (self-insert-command (prefix-numeric-value arg))
    (progn
      (delete-char -1)
      (if (= (preceding-char) ? )
          (insert "=> ")
        (insert " => ")))))



(defun vhdl-semicolon (arg)
  " ';;' --> ':='"
  (interactive "P")
  (vhdl-expand-abbrev)
  (if (or (/= (preceding-char) last-input-char)
          (vhdl-in-comment-p)
          (vhdl-in-quote-p))
      (self-insert-command (prefix-numeric-value arg))
    (progn
      (delete-char -1)
      (if (= (preceding-char) ? )
          (insert ":= ")
        (insert " := ")))))



(defun vhdl-single-quote (arg)
  "Replace to consecutive single quotes by a double quote."
  (interactive "P")
  (vhdl-expand-abbrev)
  (if (or (/= (preceding-char) last-input-char)
          (vhdl-in-comment-p))
      (self-insert-command (prefix-numeric-value arg))
    (delete-char -1)
    (insert-char ?\" 1)))

(defun vhdl-paren-close (arg)
  "if ')' is the first character on the line then vhdl-indent-line"
  (interactive "P")
  (vhdl-expand-abbrev)
  (let ((pos (point))
        (first nil))
    (insert ")")
    (save-excursion
      (back-to-indentation)
      (if (= pos (point))
          (setq first t)))
    (if first
        (vhdl-indent-line))
    (delete-char -1)
    (self-insert-command (prefix-numeric-value arg))))
          
;; set up electric character functions to work with pending-del,
;; (a.k.a. delsel) mode.
(mapcar
 (function
  (lambda (sym)
    (put sym 'delete-selection t)	; for delsel (Emacs)
    (put sym 'pending-delete t)))	; for pending-del (XEmacs)
 '(vhdl-paren-close
   vhdl-single-quote
   vhdl-period
   vhdl-comma
   vhdl-punctuation
   vhdl-newline
   vhdl-dash
   vhdl-space
   vhdl-semicolon))

;;===========================================================================;;
;;                             comment functions                             ;;
;;===========================================================================;;


(defun vhdl-comment-region (start end)
  "Comment out all lines in region using VHDL comments."
  (interactive "r")
  (let ((last (copy-marker end)))
    (goto-char start)
    (beginning-of-line)
    (while (< (point) last)
      (if (not (looking-at "\\s-*--"))   ; do not comment more than once
          (insert vhdl-comment-region-prefix))
      (beginning-of-line 2))
    (set-marker last nil)))



(defun vhdl-uncomment-region (start end)
  "Remove all leading VHDL comments in region starting with
   vhdl-comment-region-prefix string."
  (interactive "r")
  (goto-char start)
  (beginning-of-line)
  (while (re-search-forward (concat "^\\(\\s-*\\)"
						    (regexp-quote vhdl-comment-region-prefix)) end t)
   (replace-match "\\1")) ; leave only the whitespace...
  )


(defun vhdl-comment-indent ()
  (let* ((opoint (point))
         (col (progn
                (forward-line -1)
                (if (re-search-forward "--" opoint t)
                    ;; Existing comment at bol stays there.
                    (- (current-column) 2)
                  (goto-char opoint)
                  (skip-chars-backward " \t")
                  ;; Else indent to comment column
                  ;; except leave at least one space.
                  (max comment-column (1+ (current-column)))))))
    (goto-char opoint)
    col))



(defun vhdl-inline-comment ()
  "Start a comment at the end of the line.
  If on line with code, indent at least comment-column.
  If starting after vhdl-end-comment-column, start a new line."
  (interactive)
  (if (> (current-column) vhdl-end-comment-column)
      (newline-and-indent))
  (if (looking-at "\\s-*$") ; no trailing text
      (let ((margin (current-column)))
        (if (> margin comment-column)
            (progn
              (and (not (vhdl-backward-looking-at "\\s-+"))
                   (insert ? ))
              (insert "-- "))
          ;; point is still left of comment-start-column
          (delete-horizontal-space)
          (if (bolp)
              (progn
                (vhdl-indent-line)
                (insert "-- "))
            (indent-to comment-column)
            (insert "-- "))))
    ;; commenting out code following point:
    (and (not (bolp))
         (not (vhdl-backward-looking-at "\\s-+"))
         (insert ? ))
    (insert "-- ")))




(defun vhdl-block-comment ()
  "Inserts 3 comment lines at the current indent, making a block comment."
  (interactive)
  (cond
   ;; if no code between point and bol: start comment on preceeding line
   ((vhdl-first-on-line-p)
    (beginning-of-line nil)
    (insert-char ?\n 1)
    (forward-char -1)
    (vhdl-indent-line))

   ;; point not at bol: start comment on following line
   ((save-excursion
      (beginning-of-line nil)
      (not (looking-at "[ \t]*$")))
    (progn
      (end-of-line)
      (newline-and-indent)))

   (t
    (vhdl-indent-line)))

  ;; create block comment
  (let* ((col (current-column))
         (len (- vhdl-end-comment-column col)))
    (insert-char ?- len) ;;1st line of dashes
    (insert-char ?\n 1)
    (insert-char ?  col) ;;indentation
    (if (not vhdl-frame-block-comments)
	  (insert "--  \n") ;;true-sttm
	  (insert "--") ;;false-sttms
	  (insert-char ?  (- len 4))
	  (insert "--\n"))
    (insert-char ?  col) ;;indentation
    (insert-char ?- len) ;;trailing line of dashes
    (beginning-of-line 0) ;; *BACK* one line !!!!
    (forward-char (+ col 4))))


;;===========================================================================;;
;;                          VHDL language constructs                         ;;
;;===========================================================================;;


(defun vhdl-keyword-hook (callback)
  "Execute CALLBACK, but only if certain criteria are met."
  (let ((case-fold-search nil)
        (save-last-input last-input-char))
    ;; two-letter abbreviations will not be expanded unless they are in
    ;; all uppercase; exceptions are all two-letter VHDL keywords as described
    ;; in vhdl-short-keywords which may appear in any combination.
    (if (or (vhdl-in-comment-p)
            (vhdl-in-quote-p)
            (and (< (length last-abbrev-text) 3)
                 (not (string-match vhdl-short-keywords last-abbrev-text))
                 (not (string-match "^[A-Z]+$" last-abbrev-text))))
        ;; if abbrev could not be expanded it must be undone now
        (insert-string last-abbrev-text)
;;      (progn (undo-start)
;;             (undo-more 1)
;;             (forward-word 1))
      (progn
        (funcall callback)
        (if (/= last-input-char save-last-input)
            (top-level))))))

(defun vhdl-attributes-hook (callback)
  "Execute CALLBACK, but only if certain criteria are met."
  (let ((case-fold-search t)
        (save-last-input last-input-char))
    ;; two-letter abbreviations will not be expanded unless they are in
    ;; all uppercase; exceptions are all two-letter VHDL keywords as described
    ;; in vhdl-short-keywords which may appear in any combination.
    (if (and (vhdl-backward-looking-at "'\\s-*")
             (not (vhdl-in-comment-p))
             (not (vhdl-in-quote-p)))
        (funcall callback)
          ;;; push back a backspace char to delete unread ^M character
      ;; if abbrev could not be expanded it must be undone now
      (progn
        ;; RANGE must be handled in a special way as it may be both keyword
        ;; and attribute name. At this point we know it must be treated as
        ;; a keyword because it is not preceeded by a tick character:
        (cond ((eq callback 'vhdl-tick-range)
               (funcall 'vhdl-range))
              (t
               (insert-string last-abbrev-text))))
      (if (/= last-input-char save-last-input)
          (top-level)))))
;;             (undo-start)
;;             (undo-more 1)
;;             (forward-word 1)))))))




(defun vhdl-begin-end (&optional name)
  "Insert a begin ... end pair with optional NAME after the end.
   Point is left between them.
   NAME is a list with elements KEYWORD|nil IDENTIFIER INDENT."
  (let ((keyw nil))
    (if (and name (listp name))
        (progn (insert-char ?\n 1)
               (indent-to (nth 2 name))
               (setq keyw (nth 0 name))
               (setq name (nth 1 name)))
      (newline-and-indent))
    (let ((return (point))
          (margin (current-column)))
      (newline)
      (indent-to margin)
      (vhdl-insert-keyword "BEGIN") (newline)
      (indent-to (+ margin vhdl-indent))
      (newline)
      (indent-to margin)
      (vhdl-insert-keyword "END")
      (vhdl-insert-keyword (and keyw (concat " " keyw)))
      (vhdl-insert-identifier (and name (concat " " name)))
      (insert ?\;)
      (goto-char return)
      (indent-to (+ margin vhdl-indent)))))



(defun vhdl-enclose-in-parens (string)
  "Remove leading and trailing spaces from string and enclose string in
   parenthesis as needed and return the resulting string."
  ;; remove optional leading whitespace
  (and (string-match "^[ \t]*\\(.*\\)" string)
       (setq string (substring string (match-beginning 1) (match-end 1))))
  ;; remove optional trailing whitespace
  (and (string-match "\\(.*\\)\\s-+$" string)
       (setq string (substring string (match-beginning 1) (match-end 1))))

  (cond ((string-match "^[^(].*[^)]$" string)
         (concat "(" string ")"))
        ((string-match "^(.*[^)]$" string)
         (concat string ")"))
        ((string-match "^[^(].*)$" string)
         (concat "(" string ))
        (t
         string)))



(defun vhdl-insert-identifier (string)
  "Insert VHDL identifier at point according to vhdl-identifier-case-defun."
  (and string
       (insert (funcall vhdl-identifier-case-defun string))))

(defun vhdl-insert-keyword (string)
  "Insert VHDL keyword at point according to vhdl-keyword-case-defun."
  (if string
      (progn
        (insert (funcall vhdl-keyword-case-defun string))
        (if (string-match vhdl-left-indent-regexp (concat string " "))
            (vhdl-indent-line)))))

(defun vhdl-insert-attribute (string)
  "Insert VHDL attribute name at point according to vhdl-attribute-case-defun."
  ;; delete any whitespaces between tick character and attribute name
  (and string
       (progn
         (delete-horizontal-space)
         (insert (funcall vhdl-attribute-case-defun string)))))



;; hooks for all VHDL keywords:
(mapcar '(lambda (x)
           (let* ((fun  (make-symbol (concat "vhdl-" x)))
                  (fun  (intern (symbol-name fun)))
                  (hook (make-symbol (concat "vhdl-" x "-hook")))
                  (hook (intern (symbol-name hook))))
             (fset hook (` (lambda nil (vhdl-keyword-hook (quote (, fun))))))))
        vhdl-keyword-list)
      

;; hooks & functions for all attributes:
(mapcar '(lambda (x)
           (let* ((fun  (make-symbol (concat "vhdl-tick-" x)))
                  (fun  (intern (symbol-name fun)))
                  (hook (make-symbol (concat "vhdl-tick-" x "-hook")))
                  (hook (intern (symbol-name hook))))
             (fset hook (` (lambda nil (vhdl-attributes-hook (quote (, fun))))))
             (fset fun  (` (lambda nil (vhdl-insert-attribute (, x)))))))
           vhdl-attribute-list)


;; special hooks & complex keyword functions:

(defun vhdl-architecture ()
  "Insert architecture, prompting for name and related entity.
   Uses vhdl-saved-entity-name as default value if non-empty."
  (if (vhdl-looking-at-eol-p)
      (let* ((margin (current-column))
             (arch-name))
        (vhdl-insert-keyword "ARCHITECTURE ")
        (setq arch-name (vhdl-prompt "arch_name" " OF "))
        (setq vhdl-saved-architecture-name arch-name)
        (if (string-match "^\\s-*$" vhdl-saved-entity-name)
            (vhdl-prompt "ent_name" " IS")
          (vhdl-prompt (concat "ent_name" "=" vhdl-saved-entity-name) " IS"))
        (vhdl-begin-end (list nil (or arch-name "") margin))
        (indent-to (+ margin vhdl-indent)))
    (vhdl-insert-keyword "ARCHITECTURE")))

(defun vhdl-begin ()
  (vhdl-insert-keyword "BEGIN")
  (vhdl-indent-line))

(defun vhdl-block ()
  "Inserts a block statement skeleton if not preceded by text,
   in which case simply insert string BLOCK."
  (if (vhdl-looking-at-eol-p)
    (let ((margin (current-column))
          (label  (vhdl-prompt "blk_label")))
      (if (not (string-equal label ""))
          (progn
            (insert ": ")))
      (vhdl-insert-keyword "BLOCK ")
      (vhdl-prompt "expression")
      (vhdl-begin-end (list "BLOCK" (or label "") margin))
      (indent-to (+ margin vhdl-indent)))
    (vhdl-insert-keyword "BLOCK")))


(defun vhdl-case ()
  (if (vhdl-looking-at-eol-p)
    (let ((margin (current-column)))
      (vhdl-insert-keyword "CASE ")
      (vhdl-prompt "expression" " IS\n\n")
      (indent-to margin)
      (vhdl-insert-keyword "END CASE;")
      (forward-line -1)
      (indent-to (+ margin vhdl-indent))
      (vhdl-when))
    (vhdl-insert-keyword "CASE"))
  (vhdl-indent-line))


(defun vhdl-component ()
  (if (vhdl-looking-at-eol-p)
    (let ((margin (current-column)))
      (vhdl-insert-keyword "COMPONENT ")
      (vhdl-prompt "comp_name" " \n\n")
      (indent-to margin)
      (vhdl-insert-keyword "END COMPONENT;")
      (end-of-line 0)
      (vhdl-indent-line))
    (vhdl-insert-keyword "COMPONENT")))


(defun vhdl-configuration ()
  "Inserts a configuration declaration skeleton.
   Uses vhdl-saved-entity-name and vhdl-saved-architecture-name as defaults if non-empty."
  (if (vhdl-looking-at-eol-p)
      (if (save-excursion
            (forward-word -1)
            (looking-at "[Uu][Ss][Ee]"))
          (progn
            (vhdl-insert-keyword "CONFIGURATION ")
            (vhdl-prompt "lib.cfg" ";"))
        (let ((margin (current-column))
              (conf-name))
          (vhdl-insert-keyword "CONFIGURATION ")
          (setq conf-name (vhdl-prompt "cfg_name" " OF "))
          (if (string-match "^\\s-*$" vhdl-saved-entity-name)
              (vhdl-prompt "ent_name" " IS\n")
            (vhdl-prompt (concat "ent_name" "=" vhdl-saved-entity-name) " IS\n"))
          (vhdl-indent-line)
          (vhdl-insert-keyword "FOR ")
          (if (string-match "^\\s-*$" vhdl-saved-architecture-name)
              (vhdl-prompt "arch_name" "\n\n")
            (vhdl-prompt (concat "arch_name" "=" vhdl-saved-architecture-name) "\n\n"))
          (indent-to (+ margin vhdl-indent))
          (vhdl-insert-keyword "END FOR;\n")
          (indent-to margin)
          (vhdl-insert-keyword "END ")
          (vhdl-insert-identifier conf-name)
          (insert ?\;)
          (end-of-line -1)
          (vhdl-indent-line)))
    (vhdl-insert-keyword "CONFIGURATION")))


(defun vhdl-else ()
  (vhdl-insert-keyword "ELSE")
  (vhdl-indent-line))


(defun vhdl-elsif ()
  (vhdl-insert-keyword "ELSIF ")
  (vhdl-indent-line)
  (vhdl-prompt "expression" " THEN\n")
  (vhdl-indent-line))


(defun vhdl-end ()
  (vhdl-insert-keyword "END")
  (vhdl-indent-line))


(defun vhdl-endgenerate-hook ()
  (vhdl-keyword-hook 'vhdl-endgenerate))

(defun vhdl-endgenerate ()
  (vhdl-insert-keyword "END GENERATE")
  (vhdl-indent-line))


(defun vhdl-endif-hook ()
  (vhdl-keyword-hook 'vhdl-endif))

(defun vhdl-endif ()
  (vhdl-insert-keyword "END IF")
  (vhdl-indent-line))


(defun vhdl-endloop-hook ()
  (vhdl-keyword-hook 'vhdl-endloop))

(defun vhdl-endloop ()
  (vhdl-insert-keyword "END LOOP")
  (vhdl-indent-line))


(defun vhdl-endprocess-hook ()
  (vhdl-keyword-hook 'vhdl-endprocess))

(defun vhdl-endprocess ()
  (vhdl-insert-keyword "END PROCESS")
  (vhdl-indent-line))


(defun vhdl-entity ()
  "Insert entity declaration, prompting for name"
  (if (vhdl-looking-at-eol-p)
      (if (save-excursion
            (forward-word -1)
            (looking-at "[Uu][Ss][Ee]"))
          (progn
            (vhdl-insert-keyword "ENTITY ")
            (vhdl-prompt "lib.ent(arch)" ";"))
        (let* ((margin (current-column))
               (entity-name))
          (vhdl-insert-keyword "ENTITY ")
          (setq entity-name (vhdl-prompt "ent_name" " IS\n\n"))
          (setq vhdl-saved-entity-name entity-name)
          (indent-to margin)
          (vhdl-insert-keyword "END ")
          (vhdl-insert-identifier entity-name)
          (insert ?\;))
        (end-of-line 0)
        (vhdl-indent-line))
    (vhdl-insert-keyword "ENTITY")))


(defun vhdl-for-generate-hook ()
  (vhdl-keyword-hook 'vhdl-for-generate))

(defun vhdl-for-generate ()
  (let* ((margin (current-column))
         (gen-label (vhdl-prompt "label"))
         (has-name  (not (string-equal gen-label ""))))
    (if has-name (progn (insert ": ") (setq gen-label (concat " " gen-label))))
    (vhdl-insert-keyword "FOR ")
    (vhdl-prompt "range_spec" " GENERATE\n\n")
    (indent-to margin)
    (vhdl-insert-keyword "END GENERATE")
    (vhdl-insert-identifier (or gen-label ""))
    (insert ?\;)
    (end-of-line 0)
    (indent-to (+ margin vhdl-indent))))


(defun vhdl-for-loop-hook ()
  (vhdl-keyword-hook 'vhdl-for-loop))

(defun vhdl-for-loop ()
  (let* ((margin (current-column))
         (loop-label (vhdl-prompt "label"))
         (has-name  (not (string-equal loop-label ""))))
    (if has-name (progn (insert ": ") (setq loop-label (concat " " loop-label))))
    (vhdl-insert-keyword "FOR ")
    (vhdl-prompt "range_spec" " LOOP\n\n")
    (indent-to margin)
    (vhdl-insert-keyword "END LOOP")
    (vhdl-insert-identifier (or loop-label ""))
    (insert ?\;)
    (end-of-line 0)
    (indent-to (+ margin vhdl-indent))))


(defun vhdl-for-use-hook ()
  (vhdl-keyword-hook 'vhdl-for-use))

(defun vhdl-for-use ()
  (let ((margin (current-column)))
    (vhdl-insert-keyword "FOR ")
    (vhdl-prompt "inst|ALL:comp" "\n")
    (indent-to (+ margin vhdl-indent))
    (vhdl-insert-keyword "USE ")
    (if (y-or-n-p "ENTITY ? ")
        (progn
          (vhdl-insert-keyword "ENTITY ")
          (vhdl-prompt "lib.ent(arch)" ";\n"))
      (vhdl-insert-keyword "CONFIGURATION ")
      (vhdl-prompt "lib.conf" ";\n"))
    (indent-to margin)
    (vhdl-insert-keyword "END FOR;\n" )
    (indent-to margin)))


(defun vhdl-wait-until-hook ()
  (vhdl-keyword-hook 'vhdl-wait-until))

(defun vhdl-wait-until ()
  (let* ((margin (current-column))
         (clock-name "*undefined*"))
    (vhdl-insert-keyword "WAIT UNTIL ")
    (setq clock-name (vhdl-prompt "clock"))
    (vhdl-insert-attribute "'EVENT ")
	(vhdl-insert-keyword "AND ")
	(vhdl-insert-identifier clock-name)
	(insert "  = '1';\n")
    (indent-to margin)))



(defun vhdl-function ()
  "Insert frame for function body."
  (if (vhdl-looking-at-eol-p)
      (let* ((margin (current-column))
             (func-name))
        (vhdl-insert-keyword "FUNCTION ")
        (setq func-name (vhdl-prompt "func_name"))
        (vhdl-get-parameter-list)
        (vhdl-insert-keyword " RETURN ")
        (vhdl-prompt "result_type")
        (if (not (y-or-n-p "Specification (BEGIN ... END) ? "))
            (insert ";\n")
          (vhdl-insert-keyword " IS")
          (vhdl-begin-end (list nil (or func-name "") margin)))
        (indent-to (+ margin vhdl-indent)))
    (vhdl-insert-keyword "FUNCTION")))


(defun vhdl-if ()
  (if (vhdl-looking-at-eol-p)  ;; if not followed by text ...
    (let ((margin (current-column)))
      (vhdl-insert-keyword "IF ")
      (vhdl-prompt "expression" " THEN\n\n")
      (indent-to margin)
      (vhdl-insert-keyword "END IF;")
      (end-of-line 0)
      (vhdl-indent-line))
    (vhdl-insert-keyword "IF")))


(defun vhdl-if-generate-hook ()
  (vhdl-keyword-hook 'vhdl-if-generate))

(defun vhdl-if-generate ()
  (let* ((margin (current-column))
         (gen-label (vhdl-prompt "label"))
         (has-name  (not (string-equal gen-label ""))))
    (if has-name (progn (insert ": ") (setq gen-label (concat " " gen-label))))
    (vhdl-insert-keyword "IF ")
    (vhdl-prompt "condition" " GENERATE\n\n")
    (indent-to margin)
    (vhdl-insert-keyword "END GENERATE")
    (vhdl-insert-identifier (or gen-label ""))
    (insert ?\;)
    (end-of-line 0)
    (indent-to (+ margin vhdl-indent))))


(defun vhdl-loop ()
  (if (vhdl-looking-at-eol-p)
    (let ((margin (current-column))
          (loop-label (vhdl-prompt "label")))
      (if (not (string-equal loop-label ""))
          (progn
            (insert ": ")
            (setq loop-label (concat " " loop-label))))
      (vhdl-insert-keyword "LOOP\n\n")
      (indent-to margin)
      (vhdl-insert-keyword "END LOOP")
      (vhdl-insert-identifier (or loop-label ""))
      (insert ?\;)
      (end-of-line 0)
      (indent-to (+ margin vhdl-indent)))
    (vhdl-insert-keyword "LOOP")))


(defun vhdl-procedure ()
  (if (vhdl-looking-at-eol-p)
      (let* ((margin (current-column))
             (func-name))
        (vhdl-insert-keyword "PROCEDURE ")
        (setq func-name (vhdl-prompt "proc_name"))
        (vhdl-get-parameter-list)
        (if (not (y-or-n-p "Specification (BEGIN ... END) ? "))
            (insert ";\n")
          (vhdl-insert-keyword " IS")
          (vhdl-begin-end (list nil (or func-name "") margin)))
        (indent-to (+ margin vhdl-indent)))
        (vhdl-insert-keyword "PROCEDURE")))


(defun vhdl-process ()
  "Inserts a process skeleton."
  (if (vhdl-looking-at-eol-p)
      (let ((slist)
            (margin (current-column))
            (proc-name (vhdl-prompt "label")))
        (if (not (string-equal proc-name ""))
            (progn
              (insert ": ")))
        (vhdl-insert-keyword "PROCESS ")
        (setq slist (vhdl-read "sensitivity-list"))

        (if (not (string-match "^\\s-*$" slist)) ;; if not empty ...
            (progn
              ;; if parentheses are missing around slist add them:
              (setq slist (vhdl-enclose-in-parens slist))
              (vhdl-insert-identifier slist)))

        (vhdl-begin-end (list "PROCESS" (or proc-name "") margin))
        (indent-to (+ margin vhdl-indent)))
    (vhdl-insert-keyword "PROCESS")))


(defun vhdl-record ()
  (if (not (vhdl-first-on-line-p))
      (vhdl-insert-keyword "RECORD")
    (let ((margin (current-column)))
      (vhdl-insert-keyword "RECORD\n\n")
      (indent-to margin)
      (vhdl-insert-keyword "END RECORD;")
      (end-of-line 0)
      (vhdl-indent-line))))


(defun vhdl-severity ()
  (if (looking-at "[ \t]*$")
      (progn
        (vhdl-insert-keyword "SEVERITY ")
        (vhdl-prompt "note|warning|error|failure=error" ";"))
    (vhdl-insert-keyword "SEVERITY")))


(defun vhdl-units ()
  (if (not (vhdl-first-on-line-p))
      (vhdl-insert-keyword "UNITS")
    (let ((margin (current-column)))
      (vhdl-insert-keyword "UNITS\n\n")
      (indent-to margin)
      (vhdl-insert-keyword "END UNITS;")
      (end-of-line 0)
      (vhdl-indent-line))))


(defun vhdl-when ()
  (vhdl-insert-keyword "WHEN")
  (let ((ind-line t))
    (if (not (vhdl-first-on-line-p))
        (save-excursion
          (beginning-of-line)
          (if (not (looking-at "\\s-*[Ww][Hh][Ee][Nn]\\s-*"))
              (setq ind-line nil)
            (beginning-of-line 0)
            (if (looking-at "\\s-*[Cc][Aa][Ss][Ee]\\s-*")
                (setq ind-line nil)))))
    (if ind-line
        (vhdl-indent-line))))
      


(defun vhdl-while ()
  "Insert a WHILE LOOP at point asking for loop label and expression."
  (if (vhdl-looking-at-eol-p)
      (let* ((margin (current-column))
             (loop-label (vhdl-prompt "label"))
             (has-name  (not (string-equal loop-label ""))))
        (if has-name (progn (insert ": ") (setq loop-label (concat " " loop-label))))
        (vhdl-insert-keyword "WHILE ")
        (vhdl-prompt "expression" " LOOP\n\n")
        (indent-to margin)
        (vhdl-insert-keyword "END LOOP")
        (vhdl-insert-identifier (or loop-label ""))
        (insert ?\;)
        (end-of-line 0)
        (indent-to (+ margin vhdl-indent)))
    (vhdl-insert-keyword "WHILE")))

;; overwrite special hooks if vhdl-abbrev-mode is nil
(mapcar '(lambda (x)
           (let* ((fun  (make-symbol (concat "vhdl-" x)))
                  (fun  (intern (symbol-name fun))))
             (fset fun  (` (lambda nil (vhdl-insert-keyword (, x)))))))
        (if vhdl-abbrev-mode vhdl-simple-keyword-list vhdl-keyword-list))


;;===========================================================================;;
;;   search functions ignoring quoted strings and comments                   ;;
;;===========================================================================;;

(defun vhdl-re-search-forward (regexp &optional bound no-error count)
  (let ((pos (point))
        (goon t)
        (result nil)
        (data (match-data))
        (found nil))
    (while goon
      (setq result (re-search-forward regexp bound t count))
      (if result
          (progn
            (setq data (match-data))
            (if (not (or (vhdl-in-quote-p)
                         (vhdl-in-comment-p)))
                (progn
                  (setq goon nil)
                  (setq found t))))
        (setq goon nil)))
    (if (and (not found) (not no-error))
        (progn
          (goto-char pos)
          (error (concat "Search failed: \"" regexp "\""))))
    (set-match-data data)
    result))

(defun vhdl-re-search-backward (regexp &optional bound no-error count)
  (let ((pos (point))
        (goon t)
        (result nil)
        (data (match-data))
        (found nil))
    (while goon
      (setq result (re-search-backward regexp bound t count))
      (if result
          (progn
            (setq data (match-data))
            (if (not (or (vhdl-in-quote-p)
                         (vhdl-in-comment-p)))
                (progn
                  (setq goon nil)
                  (setq found t))))
        (setq goon nil)))
    (if (and (not found) (not no-error))
        (progn
          (goto-char pos)
          (error (concat "Search failed: \"" regexp "\""))))
    (set-match-data data)
    result))

(defun vhdl-search-forward (regexp &optional bound no-error count)
  (let ((pos (point))
        (goon t)
        (result nil)
        (data (match-data))
        (found nil))
    (while goon
      (setq result (search-forward regexp bound t count))
      (if result
          (progn
            (setq data (match-data))
            (if (not (or (vhdl-in-quote-p)
                         (vhdl-in-comment-p)))
                (progn
                  (setq goon nil)
                  (setq found t))))
        (setq goon nil)))
    (if (and (not found) (not no-error))
        (progn
          (goto-char pos)
          (error (concat "Search failed: \"" regexp "\""))))
    (set-match-data data)
    result))

(defun vhdl-search-backward (regexp &optional bound no-error count)
  (let ((pos (point))
        (goon t)
        (result nil)
        (data (match-data))
        (found nil))
    (while goon
      (setq result (search-backward regexp bound t count))
      (if result
          (progn
            (setq data (match-data))
            (if (not (or (vhdl-in-quote-p)
                         (vhdl-in-comment-p)))
                (progn
                  (setq goon nil)
                  (setq found t))))
        (setq goon nil)))
    (if (and (not found) (not no-error))
        (progn
          (goto-char pos)
          (error (concat "Search failed: \"" regexp "\""))))
    (set-match-data data)
    result))

;;===========================================================================;;
;;             insert default features                                       ;;
;;===========================================================================;;


(defvar vhdl-file-search-path '("." "./sources"))
(defvar vhdl-file-suffixes ".vhdl:.vhd")
(defvar vhdl-search-file-close-buffer nil)

(defun vhdl-search-file (entity)
  (interactive)
  (let* (
        (filename (locate-file entity vhdl-file-search-path vhdl-file-suffixes))
        (bufferlist (buffer-list))
        (foundbuffer nil)
        (truename)
        (fontify font-lock-auto-fontify))
    (if (not (eq filename nil))
        (setq truename (file-truename (locate-file filename vhdl-file-search-path)))
      (error (concat "Can't find file " entity "(" vhdl-file-suffixes ")")))
    (while (and bufferlist (not foundbuffer))
      (if (or
           (equal (buffer-file-name (car bufferlist)) filename)
           (equal (buffer-file-name (car bufferlist)) truename))
          (setq foundbuffer (car bufferlist)))
      (setq bufferlist (cdr bufferlist)))
    (if (not foundbuffer)
        (progn
          (setq font-lock-auto-fontify nil)
          (setq foundbuffer (find-file-noselect filename))
          (setq font-lock-auto-fontify fontify)
          (setq vhdl-search-file-close-buffer foundbuffer))
      (setq vhdl-search-file-close-buffer nil)
      (setq foundbuffer foundbuffer))))

(defun vhdl-close-buffer ()
  (if vhdl-search-file-close-buffer
      (kill-buffer vhdl-search-file-close-buffer)))

(defun vhdl-get-entity-list (entity)
  "Searches for a buffer or file named <entity>.vhdl and returns the
   entity elements in a list"
  (let ((pos (point))
        (case-save)
        (save-buffer (current-buffer))
        (found-buffer (vhdl-search-file entity))
        (found-old-pos)
        (entity-begin)
        (entity-end)
        (entity-list nil)
        )
    (set-buffer found-buffer)
    (setq case-save case-fold-search)
    (setq case-fold-search t)
    (setq found-old-pos (point))
    (goto-char (point-min))
    (if (vhdl-re-search-forward (concat "entity[ \t\n]+"
                                        entity "[ \t\n]+is[ \t\n]+") nil t)
        (progn
          (backward-char 1)
          (setq entity-begin (point))
          (vhdl-re-search-forward "[ \t\n]+end[ \t\n;]")
          (vhdl-search-backward ";")
          (forward-char 1)
          (setq entity-end (point))
          (setq entity-list (vhdl-get-port-and-generic entity-begin entity-end))
          (goto-char found-old-pos)
          (setq case-fold-search case-save)
          (vhdl-close-buffer)
          (set-buffer save-buffer))
      (goto-char pos)
      (error (concat "No ENTITY "
                     entity
                     " found in "
                     (buffer-file-name (current-buffer)))))
    (goto-char pos)
    entity-list))

(defun vhdl-get-component-list (entity)
  "Searches for a buffer or file named <entity>.vhdl and returns the
   entity elements in a list"
  (let (
        (case-fold-search t)
        (old-pos (point))
        (comp-begin)
        (comp-end)
        (comp-list nil)
        )
    (setq old-pos (point))
    (goto-char (point-min))
    (if (vhdl-re-search-forward (concat
                                 "component[ \t\n]+"
                                 entity "[ \t\n]") nil t)
        (progn
          (backward-char 1)
          (setq comp-begin (point))
          (vhdl-re-search-forward "[ \t\n]+end[ \t\n;]")
          (vhdl-search-backward ";")
          (forward-char 1)
          (setq comp-end (point))
          (setq comp-list (vhdl-get-port-and-generic comp-begin comp-end))))
    (goto-char old-pos)
    comp-list))
      
(defun vhdl-get-port-and-generic (begin-point end-point)
  (let ((case-fold-search t)
        (begin 0)
        (end 0)
        (port nil)
        (generic nil))
    (goto-char begin-point)
    (if (vhdl-re-search-forward "[ \t\n]+port[ \t\n]*(" end-point t)
        (progn
          (backward-char 1)
          (setq begin (1+ (point)))
          (forward-list 1)
          (backward-char 2)
          (while (looking-at "[\t\n ]")
            (backward-char 1))
          (setq end (1+ (point)))
          (setq port (buffer-substring begin end))))
    (goto-char begin-point)
    (if (vhdl-re-search-forward "[ \t\n]+generic[ \t\n]*(" end-point t)
        (progn
          (backward-char 1)
          (setq begin (1+ (point)))
          (forward-list 1)
          (backward-char 2)
          (while (looking-at "[\t\n ]")
            (backward-char 1))
          (setq end (1+ (point)))
          (setq generic (buffer-substring begin end))))
    (list generic port)))

(defun vhdl-get-signal-names (string)
  (let ((sig-list nil)
        (signal-str (replace-in-string string "\\(\\s-*--.*\\)" "")))
    (setq signal-str (replace-in-string signal-str "\\([ \t\n]*:[^;]*\\)" ""))
    (setq signal-str (replace-in-string signal-str "\\([ \t\n,;]+\\)" " "))
    (while (string-match "[^ ]+" signal-str)
      (setq sig-list (append sig-list (list (substring signal-str
                                                       (match-beginning 0)
                                                       (match-end 0)))))
      (setq signal-str (substring signal-str (match-end 0))))
    (setq sig-list sig-list)))


(defun vhdl-insert-default-port-map ()
  (interactive)
  (let* ((pos (point))
         (pm-begin)
         (label (read-string "Instantiation Label: " nil nil))
         (entity (read-string "Entity Name: " nil nil))
         (entity-list nil)
         (generic-list nil)
         (port-list nil)
         (str ""))
    (beginning-of-line)
    (setq pm-begin (point))
    (goto-char pos)
    (if (not (setq entity-list (vhdl-get-component-list entity)))
        (setq entity-list (vhdl-get-entity-list entity)))
    (setq str (concat label ": " entity "\n"))
    (if (car entity-list)
        (progn
          (setq generic-list (vhdl-get-signal-names (car entity-list)))
          (setq str (concat str "generic map (\n"))
          (while generic-list
            (setq str (concat str (car generic-list) " => " (car generic-list)))
            (setq generic-list (cdr generic-list))
            (if generic-list
                (setq str (concat str ",\n"))
              (setq str (concat str "\n"))))
          (setq str (concat str ")\n"))))
    (if (car (cdr entity-list))
        (progn
          (setq port-list (vhdl-get-signal-names (car (cdr entity-list))))
          (setq str (concat str "port map (\n"))
          (while port-list
            (setq str (concat str (car port-list) " => " (car port-list)))
            (setq port-list (cdr port-list))
            (if port-list
                (setq str (concat str ",\n"))
              (setq str (concat str "\n"))))
          (setq str (concat str ")\n"))))
    (goto-char pos)
    (insert (concat str ";"))
    (vhdl-vfmt-region pm-begin (point))
    (goto-char pos)))


(defun vhdl-replace-comma-list (comma-list)
  (if comma-list
      (progn
        (let ((pos (point))
              (type nil))
          (save-excursion
            (vhdl-search-backward ":")
            (forward-sexp 1)
            (backward-sexp 1)
            (setq type (buffer-substring (point) pos))
            (while comma-list
              (goto-char (car comma-list))
              (delete-char 1)
              (insert (concat " : " type ";"))
              (forward-sexp 1)
              (backward-sexp 1)
              (insert "signal ")
              (setq comma-list (cdr comma-list))))))))

(defun vhdl-insert-default-signals ()
  (interactive)
  (let* ((pos (point))
         (end nil)
         (regexp "[,;:]")
         (pm-begin)
         (entity (read-string "Entity Name: " nil nil))
         (entity-list nil)
         (comma-list nil)
         (str ""))
    (beginning-of-line)
    (setq pm-begin (point))
    (goto-char pos)
    (if (not (setq entity-list (vhdl-get-component-list entity)))
        (setq entity-list (vhdl-get-entity-list entity)))
    (if (car (cdr entity-list))
        (progn
          (setq str
                (replace-in-string (car (cdr entity-list)) "\\([ \t\n]*:=[^;|\\(--\\)]*\\)" ""))
          (setq str
                (replace-in-string str
                                   ":[\t\n ]*\\(inout\\|out\\|in\\|buffer\\|linkage\\)"
                                   ":"))
          (if (looking-at "[ \t]*[^\n]")
              (progn
                (end-of-line)
                (newline-and-indent)
                (setq pos (point))))
          (if (eobp)
              (progn
                (insert "\n")
                (backward-char 1)))
          (insert str)
          (setq end (- (point-max) (point)))
          (goto-char pos)
          (if (looking-at "[\t\n ]*\\(--.*[\t\n ]*\\)*")
            (kill-region (match-beginning 0) (match-end 0)))
          (forward-sexp 1)
          (backward-sexp 1)
          (insert "signal ")
          (while (vhdl-re-search-forward regexp (- (point-max) end) t)
            (backward-char 1)
            (cond
             ((looking-at ";")
              (setq regexp "[,:]")
              (vhdl-replace-comma-list comma-list)
              (setq comma-list nil)
              (forward-sexp 1)
              (backward-sexp 1)
              (insert "signal "))
             ((looking-at ",")
              (setq comma-list (cons (point) comma-list))
              (forward-char 1))
             ((looking-at ":")
              (setq regexp ";"))))
          (goto-char (- (point-max) end))
          (forward-char 1)
          (backward-sexp 1)
          (forward-sexp 1)
          (vhdl-replace-comma-list comma-list)
          (insert ";")
          ))
    (goto-char (- (point-max) end))
    (end-of-line)
    (forward-char 1)
    (vhdl-vfmt-region pm-begin (point))
    (goto-char pos)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; vhdl-insert-default-component
;; prompts for an entity name, reads the corresponding entity descrition
;; and constructs the corresponding component declaration.
;; 
;;
;; the list is ordered: first architecture comes first
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun vhdl-insert-default-component ()
  (interactive)
  (let ((pos (point))
        (entity (read-string "Entity Name: " nil nil))
        (entity-list nil)
        (entity-str)
        (comp-begin))
    (beginning-of-line)
    (setq comp-begin (point))
    (goto-char pos)
    (setq entity-list (vhdl-get-entity-list entity))
    (setq entity-str (concat "component " entity "\n"))
    (if (car entity-list)
        (setq entity-str
			  ;;kick out the default value assignments
              (concat entity-str "generic ("
				  (replace-in-string (car entity-list)
						"\\([ \t\n]*:=[^;]*?\\(;\\|$\\)\\)" "\\2")
                      "\n);")))
				;;(concat entity-str "\npopel (\n" (car entity-list) "\npopel)\n")))
    (if (car (cdr entity-list))
        (setq entity-str
              (concat entity-str "port ("
                      (replace-in-string (car (cdr entity-list))
                                         "\\([ \t\n]*:=[^;|\\(--\\)]*\\)" "")
                      "\n);")))
    (setq entity-str (concat entity-str "\nend component;"))
    (insert entity-str)
    (vhdl-vfmt-region comp-begin (point))
    (goto-char pos)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; vhdl-search-architecture returns a list of all architectures that
;; could be found for a given entity name. Each list entry represents
;; one architecture, as a list consisiting of
;; - architecture name (string)
;; - position of architecture KW (point)
;; - position of begin (point)
;; - position of end (point)
;;
;; the list is ordered: first architecture comes first
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun vhdl-search-architecture (entity)
  (let ((case-fold-search t)
        (first)
        (begin)
        (end)
        (arch)
        (return-list nil))
    (while (vhdl-re-search-forward (concat
                               "[ \t\n]architecture[ \t\n]+\\([a-zA-Z0-9_]+\\)[ \t\n]+of[ \t\n]+"
                               entity
                               "[ \t\n]+is[ \t\n]") nil t)
      (setq arch (buffer-substring (match-beginning 1) (match-end 1)))
      (setq first (point))
      (backward-char 1)
      (if (not (vhdl-re-search-forward "[ \t\n]begin[ \t\n]" nil t))
          (error "No Keyword BEGIN found")
        (setq begin (point))
        (if (not (vhdl-re-search-forward (concat "[ \t\n]end[ \t\n]+"
                                                 arch
                                                 "[ \t\n]*;") nil t))
            (error (concat "No END " arch "; found"))
          (setq end (match-beginning 0))
          (setq return-list (cons (list arch first begin end) return-list)))))
    (reverse return-list)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; vhdl-get-architecture-hierarchy
;; extracts the block hierarchy within one architecture's begin/end block (defined
;; by the start/end points). The hierarchy is represented as list-of lists (sorted 
;; by ascending line numbers, each list entry representin an open/close 
;; block/generate/component instantiation
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;this one's a qsort callback that will sort the itmes found below to adhere
;; ascending line numbers.
(defun vhdl-sort-hierarchy (item1 item2)
  (< (car (cdr item1)) (car (cdr item2))))


(defun vhdl-get-architecture-hierarchy (start end)
  (interactive)
  (let ((case-fold-search t)
        (hier-list '())
        (gen-tmp start)
        (gen-beg))
	;; at first, we create a list of all block sttm headers (rev order).
	; each one is represented by a sublist:
	; - "block-begin" tag
	; - char pos after the BLOCK KW (position)
	; - block label (string)
	; - begin and end pos of the label
    (goto-char start)
    (while (vhdl-re-search-forward "[ \t\n]\\([a-zA-Z0-9_]+\\)[ \t\n]*:[ \t\n]*block[ \t\n]" end t)
      (setq hier-list (cons (list "block-begin" (match-end 0) (buffer-substring (match-beginning 1) (match-end 1))) hier-list)))
	
	;;now, let's add the block footers:
	; "block-end" label
	; position before the END keyword
    (goto-char start)
    (while (vhdl-re-search-forward "[ \t\n]end[ \t\n]+block\\([ \t\n]\\|;\\)" end t)
      (setq hier-list (cons (list "block-end" (match-beginning 0)) hier-list)))
      
	 ;; capture the generate sttms
	 ; - "gener-begin" label
	 ; - position after the generate KW
	 ; - generate label (string)
    (goto-char start)
    (while (vhdl-re-search-forward "[ \t\n]generate[ \t\n]+[^;]" end t)
      (setq gen-beg (- (match-end 0) 2))
      (if (vhdl-re-search-backward "[ \t\n]\\([a-zA-Z0-9_]+\\)[ \t\n]*:[ \t\n]*\\(if\\|for\\)" gen-tmp t)
          (setq hier-list (cons (list "gener-begin" gen-beg (buffer-substring (match-beginning 1) (match-end 1))) hier-list)))
      (setq gen-tmp gen-beg)
      (goto-char gen-beg))
      
	 ;; capture the end generates
	 ; - "gener-end" label
	 ; - position just before the END KW
    (goto-char start)
    (while (vhdl-re-search-forward "[ \t\n]end[ \t\n]+generate\\([ \t\n]\\|;\\)" end t)
      (setq hier-list (cons (list "gener-end" (match-beginning 0)) hier-list)))
      
	;;capture the component instantiations
	; - "instances" label
	; - position before instance name label
	; - instance name (string)
	; - component name (string)
    (goto-char start)
    (while (vhdl-re-search-forward "[ \t\n]\\([a-zA-Z0-9_]+\\)[ \t\n]*:[ \t\n]*\\([a-zA-Z0-9_]+\\)\\(\\(--.*?\n\\)\\|[ \t\n]\\)+\\(generic\\|port\\)[ \t\n]+map"
                                   end t)
      (setq hier-list (cons (list "instances"
                                  (match-beginning 1)
                                  (buffer-substring (match-beginning 1) (match-end 1))
                                  (buffer-substring (match-beginning 2) (match-end 2))) hier-list)))
    (sort hier-list 'vhdl-sort-hierarchy)))



(defun vhdl-check-for-multiple-components (item-list)
  (let ((i-list nil)
        (item)
        (cmp-item)
        (count)
        (entity))
    (while item-list
      (setq item (car item-list))
      (if (equal "instances" (car item))
          (progn
            (setq count 1)
            (setq cmp-item (nth count item-list))
            (setq entity (car (cdr (cdr (cdr item)))))
            (while cmp-item
              (cond ((equal "instances" (car cmp-item))
                     (if (equal entity (car (cdr (cdr (cdr cmp-item)))))
                         (progn
                           (if (not (cdr (cdr (cdr (cdr item)))))
                               (progn
                                 (setcar item-list (append item (list (concat "-- " (car (cdr (cdr item)))))))
                                 (setq item (car item-list))))
                           (setcar (cdr (cdr (cdr (cdr item)))) (concat (car (cdr (cdr (cdr (cdr item))))) ", " (car (cdr (cdr cmp-item)))))
                           (setcar (cdr (cdr item)) "all")
                           (setcdr (nthcdr (1- count) item-list) (nthcdr (1+ count) item-list)))
;                          (setcdr item-list (nthcdr (1+ count) item-list)))
                       (setq count (1+ count)))
                     (setq cmp-item (nth count item-list)))
                    ((equal "block-begin" (car cmp-item))
                     (setq cmp-item nil))
                    ((equal "block-end" (car cmp-item))
                     (setq cmp-item nil))
                    ((equal "gener-begin" (car cmp-item))
                     (setq cmp-item nil))
                    ((equal "gener-end" (car cmp-item))
                     (setq cmp-item nil))))))
      (setq i-list (cons (car item-list) i-list))
      (setq item-list (cdr item-list)))
    (reverse i-list)))



(defun vhdl-insert-default-configuration ()
  (interactive)
  (let ((pos (point))
        (pos-mark (point))
        (arch-list)
        (arch-item)
        (item-list)
        (arch-name) (arch-begin) (arch-end)
        (arch-count 1) (arch-count-str "")
        (item)
        (entity (read-string "Entity Name: " nil nil))
        )
    (goto-char (point-min))
    (if (not (setq arch-list (vhdl-search-architecture entity)))
        (error (concat "No Architecture for " entity " found")))

    (while arch-list ;for each architecture in arch-list
      (setq arch-item (car arch-list))
      (setq arch-list (cdr arch-list))
      (setq arch-name  (first arch-item))
      (setq arch-begin (third arch-item))
      (setq arch-end   (fourth arch-item))
      (if (> arch-count 1)
          (setq arch-count-str (concat "_" (number-to-string arch-count))))
      (setq item-list (vhdl-get-architecture-hierarchy arch-begin arch-end))
      (setq item-list (vhdl-check-for-multiple-components item-list))
      (goto-char pos-mark)
      (insert (concat "configuration " entity arch-count-str vhdl-configuration-suffix " of " entity " is\nfor " arch-name "\n"))
      (while item-list
        (setq item (car item-list))
        (cond ((equal "instances" (car item))
               (insert (concat "for " (car (cdr (cdr item))) ": " (car (cdr (cdr (cdr item)))) " " (car (cdr (cdr (cdr (cdr item))))) "\n"))
               (insert (concat "use configuration work." (car (cdr (cdr (cdr item)))) vhdl-configuration-suffix ";\n"))
               (insert "end for;\n"))
              ((equal "block-begin" (car item))
               (insert (concat "for " (car (cdr (cdr item))) "\n")))
              ((equal "block-end" (car item))
               (insert "end for;\n"))
              ((equal "gener-begin" (car item))
               (insert (concat "for " (car (cdr (cdr item))) "\n")))
              ((equal "gener-end" (car item))
               (insert "end for;\n")))
        (setq item-list (cdr item-list)))
      (insert (concat "end for;\nend " entity arch-count-str vhdl-configuration-suffix ";\n\n"))
      (setq arch-count (1+ arch-count))
      (setq pos-mark (point)))
    (vhdl-vfmt-region pos (point))
    (goto-char pos)))
     

;;============================================================================;;
;;                                  load hooks                                ;;
;;============================================================================;;

(run-hooks 'vhdl-load-hook)


;;============================================================================;;
;;                               provided features                            ;;
;;============================================================================;;

(provide 'vhdl-mode)



;; vhdl-mode.el ends here
